from dataclasses import dataclass

import click

from unstructured_ingest.v2.cli.base import DestCmd
from unstructured_ingest.v2.cli.interfaces import CliConfig
from unstructured_ingest.v2.processes.connectors.milvus import CONNECTOR_TYPE


@dataclass
class MilvusCliConnectionConfig(CliConfig):
    @staticmethod
    def get_cli_options() -> list[click.Option]:
        options = [
            click.Option(
                ["--uri"],
                required=False,
                type=str,
                default=None,
                help="Milvus uri, eg 'http://localhost:19530",
            ),
            click.Option(
                ["--user"],
                required=False,
                type=str,
                default=None,
                help="Milvus user",
            ),
            click.Option(
                ["--password"],
                required=False,
                type=str,
                default=None,
                help="Milvus password",
            ),
            click.Option(
                ["--db-name"],
                required=False,
                type=str,
                default=None,
                help="Milvus database name",
            ),
        ]
        return options


@dataclass
class MilvusCliUploaderConfig(CliConfig):
    @staticmethod
    def get_cli_options() -> list[click.Option]:
        options = [
            click.Option(
                ["--collection-name"],
                required=True,
                type=str,
                help="Milvus collections to write to",
            ),
            click.Option(
                ["--num-of-processes"],
                type=click.IntRange(min=1),
                default=4,
                help="number of processes to use when writing to support parallel writes",
            ),
        ]
        return options


milvus_dest_cmd = DestCmd(
    cmd_name=CONNECTOR_TYPE,
    connection_config=MilvusCliConnectionConfig,
    uploader_config=MilvusCliUploaderConfig,
)



import typing as t
from dataclasses import dataclass

from unstructured_ingest.interfaces import BaseDestinationConnector
from unstructured_ingest.runner.writers.base_writer import Writer

if t.TYPE_CHECKING:
    from unstructured_ingest.connector.azure_cognitive_search import (
        AzureCognitiveSearchWriteConfig,
        SimpleAzureCognitiveSearchStorageConfig,
    )


@dataclass
class AzureCognitiveSearchWriter(Writer):
    connector_config: "SimpleAzureCognitiveSearchStorageConfig"
    write_config: "AzureCognitiveSearchWriteConfig"

    def get_connector_cls(self) -> t.Type[BaseDestinationConnector]:
        from unstructured_ingest.connector.azure_cognitive_search import (
            AzureCognitiveSearchDestinationConnector,
        )

        return AzureCognitiveSearchDestinationConnector



from __future__ import annotations

import os
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Generator, Optional
from urllib.parse import urlparse

from unstructured_ingest.enhanced_dataclass import enhanced_field
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.interfaces import DownloadResponse, FileData, UploadContent
from unstructured_ingest.v2.processes.connector_registry import (
    DestinationRegistryEntry,
    SourceRegistryEntry,
)
from unstructured_ingest.v2.processes.connectors.fsspec.fsspec import (
    FsspecAccessConfig,
    FsspecConnectionConfig,
    FsspecDownloader,
    FsspecDownloaderConfig,
    FsspecIndexer,
    FsspecIndexerConfig,
    FsspecUploader,
    FsspecUploaderConfig,
)

CONNECTOR_TYPE = "sftp"


@dataclass
class SftpIndexerConfig(FsspecIndexerConfig):
    def __post_init__(self):
        super().__post_init__()
        _, ext = os.path.splitext(self.remote_url)
        parsed_url = urlparse(self.remote_url)
        if ext:
            self.path_without_protocol = Path(parsed_url.path).parent.as_posix().lstrip("/")
        else:
            self.path_without_protocol = parsed_url.path.lstrip("/")


@dataclass
class SftpAccessConfig(FsspecAccessConfig):
    password: str


@dataclass
class SftpConnectionConfig(FsspecConnectionConfig):
    supported_protocols: list[str] = field(default_factory=lambda: ["sftp"])
    access_config: SftpAccessConfig = enhanced_field(sensitive=True)
    connector_type: str = CONNECTOR_TYPE
    username: Optional[str] = None
    host: Optional[str] = None
    port: int = 22
    look_for_keys: bool = False
    allow_agent: bool = False

    def get_access_config(self) -> dict[str, Any]:
        access_config = {
            "username": self.username,
            "host": self.host,
            "port": self.port,
            "look_for_keys": self.look_for_keys,
            "allow_agent": self.allow_agent,
            "password": self.access_config.password,
        }
        return access_config


@dataclass
class SftpIndexer(FsspecIndexer):
    connection_config: SftpConnectionConfig
    index_config: SftpIndexerConfig
    connector_type: str = CONNECTOR_TYPE

    @requires_dependencies(["paramiko", "fsspec"], extras="sftp")
    def __post_init__(self):
        parsed_url = urlparse(self.index_config.remote_url)
        self.connection_config.host = parsed_url.hostname or self.connection_config.host
        self.connection_config.port = parsed_url.port or self.connection_config.port

    @requires_dependencies(["paramiko", "fsspec"], extras="sftp")
    def run(self, **kwargs: Any) -> Generator[FileData, None, None]:
        for file in super().run(**kwargs):
            new_identifier = (
                f"sftp://"
                f"{self.connection_config.host}:"
                f"{self.connection_config.port}/"
                f"{file.identifier}"
            )
            file.identifier = new_identifier
            yield file


@dataclass
class SftpDownloaderConfig(FsspecDownloaderConfig):
    remote_url: Optional[str] = None

    def __post_init__(self):
        # TODO once python3.9 no longer supported and kw_only is allowed in dataclasses, remove:
        if not self.remote_url:
            raise TypeError(
                f"{self.__class__.__name__}.__init__() "
                f"missing 1 required positional argument: 'remote_url'"
            )


@dataclass
class SftpDownloader(FsspecDownloader):
    protocol: str = "sftp"
    connection_config: SftpConnectionConfig
    connector_type: str = CONNECTOR_TYPE
    download_config: Optional[SftpDownloaderConfig] = field(default_factory=SftpDownloaderConfig)

    @requires_dependencies(["paramiko", "fsspec"], extras="sftp")
    def __post_init__(self):
        parsed_url = urlparse(self.download_config.remote_url)
        self.connection_config.host = parsed_url.hostname or self.connection_config.host
        self.connection_config.port = parsed_url.port or self.connection_config.port

    @requires_dependencies(["paramiko", "fsspec"], extras="sftp")
    def run(self, file_data: FileData, **kwargs: Any) -> DownloadResponse:
        return super().run(file_data=file_data, **kwargs)

    @requires_dependencies(["paramiko", "fsspec"], extras="sftp")
    async def run_async(self, file_data: FileData, **kwargs: Any) -> DownloadResponse:
        return await super().run_async(file_data=file_data, **kwargs)


@dataclass
class SftpUploaderConfig(FsspecUploaderConfig):
    pass


@dataclass
class SftpUploader(FsspecUploader):
    connector_type: str = CONNECTOR_TYPE
    connection_config: SftpConnectionConfig
    upload_config: SftpUploaderConfig = field(default=None)

    @requires_dependencies(["paramiko", "fsspec"], extras="sftp")
    def __post_init__(self):
        super().__post_init__()

    @requires_dependencies(["paramiko", "fsspec"], extras="sftp")
    def run(self, contents: list[UploadContent], **kwargs: Any) -> None:
        return super().run(contents=contents, **kwargs)

    @requires_dependencies(["paramiko", "fsspec"], extras="sftp")
    async def run_async(self, path: Path, file_data: FileData, **kwargs: Any) -> None:
        return await super().run_async(path=path, file_data=file_data, **kwargs)


sftp_source_entry = SourceRegistryEntry(
    indexer=SftpIndexer,
    indexer_config=SftpIndexerConfig,
    downloader=SftpDownloader,
    downloader_config=SftpDownloaderConfig,
    connection_config=SftpConnectionConfig,
)

sftp_destination_entry = DestinationRegistryEntry(
    uploader=SftpUploader,
    uploader_config=SftpUploaderConfig,
    connection_config=SftpConnectionConfig,
)



from abc import ABC
from copy import copy
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any

from unstructured_ingest.enhanced_dataclass import EnhancedDataClassJsonMixin
from unstructured_ingest.utils.compression import TAR_FILE_EXT, ZIP_FILE_EXT, uncompress_file
from unstructured_ingest.v2.interfaces import FileData
from unstructured_ingest.v2.interfaces.process import BaseProcess


@dataclass
class UncompressConfig(EnhancedDataClassJsonMixin):
    pass


@dataclass
class Uncompressor(BaseProcess, ABC):
    config: UncompressConfig = field(default_factory=UncompressConfig)

    def is_async(self) -> bool:
        return True

    def run(self, file_data: FileData, **kwargs: Any) -> list[FileData]:
        local_filepath = Path(file_data.source_identifiers.fullpath)
        if local_filepath.suffix not in TAR_FILE_EXT + ZIP_FILE_EXT:
            return [file_data]
        new_path = uncompress_file(filename=str(local_filepath))
        new_files = [i for i in Path(new_path).rglob("*") if i.is_file()]
        responses = []
        for f in new_files:
            new_file_data = copy(file_data)
            new_file_data.source_identifiers.fullpath = str(f)
            if new_file_data.source_identifiers.rel_path:
                new_file_data.source_identifiers.rel_path = str(f).replace(
                    str(local_filepath.parent), ""
                )[1:]
            responses.append(new_file_data)
        return responses

    async def run_async(self, file_data: FileData, **kwargs: Any) -> list[FileData]:
        return self.run(file_data=file_data, **kwargs)



import os
import shutil
from pathlib import Path

from unstructured_ingest.connector.registry import create_ingest_doc_from_dict
from unstructured_ingest.logger import logger
from unstructured_ingest.pipeline.interfaces import CopyNode


class Copier(CopyNode):
    def run(self, json_path: str):
        filename = os.path.basename(json_path)
        doc_hash = os.path.splitext(filename)[0]
        ingest_doc_dict = self.pipeline_context.ingest_docs_map[doc_hash]
        ingest_doc = create_ingest_doc_from_dict(ingest_doc_dict)
        desired_output = ingest_doc._output_filename
        Path(desired_output).parent.mkdir(parents=True, exist_ok=True)
        logger.info(f"Copying {json_path} -> {desired_output}")
        shutil.copy(json_path, desired_output)



import json
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from time import time
from typing import TYPE_CHECKING, Any, Generator, Optional
from urllib.parse import quote

from unstructured.documents.elements import DataSourceMetadata

from unstructured_ingest.enhanced_dataclass import EnhancedDataClassJsonMixin, enhanced_field
from unstructured_ingest.error import SourceConnectionNetworkError
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Downloader,
    DownloaderConfig,
    DownloadResponse,
    FileData,
    Indexer,
    IndexerConfig,
    SourceIdentifiers,
    download_responses,
)
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.processes.connector_registry import (
    SourceRegistryEntry,
)

from .utils import parse_datetime

if TYPE_CHECKING:
    from office365.graph_client import GraphClient
    from office365.onedrive.driveitems.driveItem import DriveItem
    from office365.onedrive.drives.drive import Drive
    from office365.onedrive.permissions.permission import Permission
    from office365.onedrive.sites.site import Site
    from office365.sharepoint.client_context import ClientContext
    from office365.sharepoint.files.file import File
    from office365.sharepoint.folders.folder import Folder
    from office365.sharepoint.publishing.pages.page import SitePage

CONNECTOR_TYPE = "sharepoint"

MAX_MB_SIZE = 512_000_000

# TODO handle other data types possible from Sharepoint
# exampled: https://github.com/vgrem/Office365-REST-Python-Client/tree/master/examples/sharepoint


class SharepointContentType(Enum):
    DOCUMENT = "document"
    SITEPAGE = "site_page"
    LIST = "list"


@dataclass
class SharepointAccessConfig(AccessConfig):
    client_cred: str


@dataclass
class SharepointPermissionsConfig(EnhancedDataClassJsonMixin):
    permissions_application_id: str
    permissions_tenant: str
    permissions_client_cred: str = enhanced_field(sensitive=True)
    authority_url: Optional[str] = field(repr=False, default="https://login.microsoftonline.com")


@dataclass
class SharepointConnectionConfig(ConnectionConfig):
    client_id: str
    site: str
    access_config: SharepointAccessConfig = enhanced_field(sensitive=True)
    permissions_config: Optional[SharepointPermissionsConfig] = None

    @requires_dependencies(["office365"], extras="sharepoint")
    def get_client(self) -> "ClientContext":
        from office365.runtime.auth.client_credential import ClientCredential
        from office365.sharepoint.client_context import ClientContext

        try:
            credentials = ClientCredential(self.client_id, self.access_config.client_cred)
            site_client = ClientContext(self.site).with_credentials(credentials)
        except Exception as e:
            logger.error(f"Couldn't set Sharepoint client: {e}")
            raise e
        return site_client

    @requires_dependencies(["msal"], extras="sharepoint")
    def get_permissions_token(self):
        from msal import ConfidentialClientApplication

        try:
            app = ConfidentialClientApplication(
                authority=f"{self.permissions_config.authority_url}/"
                f"{self.permissions_config.permissions_tenant}",
                client_id=self.permissions_config.permissions_application_id,
                client_credential=self.permissions_config.permissions_client_cred,
            )
            token = app.acquire_token_for_client(scopes=["https://graph.microsoft.com/.default"])
        except ValueError as exc:
            logger.error("Couldn't set up credentials for Sharepoint")
            raise exc
        if "error" in token:
            raise SourceConnectionNetworkError(
                "failed to fetch token, {}: {}".format(token["error"], token["error_description"])
            )
        return token

    @requires_dependencies(["office365"], extras="sharepoint")
    def get_permissions_client(self) -> Optional["GraphClient"]:
        from office365.graph_client import GraphClient

        if self.permissions_config is None:
            return None

        client = GraphClient(self.get_permissions_token)
        return client


@dataclass
class SharepointIndexerConfig(IndexerConfig):
    path: Optional[str] = None
    recursive: bool = False
    omit_files: bool = False
    omit_pages: bool = False
    omit_lists: bool = False


@dataclass
class SharepointIndexer(Indexer):
    connection_config: SharepointConnectionConfig
    index_config: SharepointIndexerConfig = field(default_factory=lambda: SharepointIndexerConfig())

    def list_files(self, folder: "Folder", recursive: bool = False) -> list["File"]:
        if not recursive:
            folder.expand(["Files"]).get().execute_query()
            return folder.files

        folder.expand(["Files", "Folders"]).get().execute_query()
        files: list["File"] = list(folder.files)
        folders: list["Folder"] = list(folder.folders)
        for f in folders:
            if "/Forms" in f.serverRelativeUrl:
                continue
            files.extend(self.list_files(f, recursive))
        return files

    def get_properties(self, raw_properties: dict) -> dict:
        raw_properties = {k: v for k, v in raw_properties.items() if v}
        filtered_properties = {}
        for k, v in raw_properties.items():
            try:
                json.dumps(v)
                filtered_properties[k] = v
            except TypeError:
                pass
        return filtered_properties

    def list_pages(self, client: "ClientContext") -> list["SitePage"]:
        pages = client.site_pages.pages.get().execute_query()
        return pages

    def page_to_file_data(self, site_page: "SitePage") -> FileData:
        site_page.expand(site_page.properties.keys()).get().execute_query()
        version = site_page.properties.get("Version", None)
        unique_id = site_page.properties.get("UniqueId", None)
        modified_date = site_page.properties.get("Modified", None)
        url = site_page.properties.get("AbsoluteUrl", None)
        date_modified_dt = parse_datetime(modified_date) if modified_date else None
        date_created_at = (
            parse_datetime(site_page.first_published)
            if (site_page.first_published and site_page.first_published != "0001-01-01T08:00:00Z")
            else None
        )
        file_path = site_page.get_property("Url", "")
        server_path = file_path if file_path[0] != "/" else file_path[1:]
        additional_metadata = self.get_properties(raw_properties=site_page.properties)
        additional_metadata["sharepoint_content_type"] = SharepointContentType.SITEPAGE.value
        return FileData(
            identifier=unique_id,
            connector_type=CONNECTOR_TYPE,
            source_identifiers=SourceIdentifiers(
                filename=site_page.file_name,
                fullpath=file_path,
                rel_path=file_path.replace(self.index_config.path, ""),
            ),
            metadata=DataSourceMetadata(
                url=url,
                version=version,
                date_modified=str(date_modified_dt.timestamp()) if date_modified_dt else None,
                date_created=str(date_created_at.timestamp()) if date_created_at else None,
                date_processed=str(time()),
                record_locator={
                    "server_path": server_path,
                },
            ),
            additional_metadata=additional_metadata,
        )

    def file_to_file_data(self, client: "ClientContext", file: "File") -> FileData:
        file.expand(file.properties.keys()).get().execute_query()
        absolute_url = f"{client.base_url}{quote(file.serverRelativeUrl)}"
        date_modified_dt = (
            parse_datetime(file.time_last_modified) if file.time_last_modified else None
        )

        date_created_at = parse_datetime(file.time_created) if file.time_created else None
        additional_metadata = self.get_properties(raw_properties=file.properties)
        additional_metadata["sharepoint_content_type"] = SharepointContentType.DOCUMENT.value
        fullpath = str(file.serverRelativeUrl)
        rel_path = fullpath.replace(self.index_config.path, "")
        while rel_path[0] == "/":
            rel_path = rel_path[1:]
        return FileData(
            identifier=file.unique_id,
            connector_type=CONNECTOR_TYPE,
            source_identifiers=SourceIdentifiers(
                filename=file.name,
                fullpath=fullpath,
                rel_path=rel_path,
            ),
            metadata=DataSourceMetadata(
                url=absolute_url,
                version=f"{file.major_version}.{file.minor_version}",
                date_modified=str(date_modified_dt.timestamp()) if date_modified_dt else None,
                date_created=str(date_created_at.timestamp()) if date_created_at else None,
                date_processed=str(time()),
                record_locator={"server_path": file.serverRelativeUrl, "site_url": client.base_url},
            ),
            additional_metadata=additional_metadata,
        )

    def get_root(self, client: "ClientContext") -> "Folder":
        if path := self.index_config.path:
            return client.web.get_folder_by_server_relative_path(path)
        default_document_library = client.web.default_document_library()
        root_folder = default_document_library.root_folder
        root_folder = root_folder.get().execute_query()
        self.index_config.path = root_folder.name
        return root_folder

    def get_site_url(self, client: "ClientContext") -> str:
        res = client.web.get().execute_query()
        return res.url

    def get_site(self, permissions_client: "GraphClient", site_url) -> "Site":
        return permissions_client.sites.get_by_url(url=site_url).execute_query()

    def get_permissions_items(self, site: "Site") -> list["DriveItem"]:
        # TODO find a way to narrow this search down by name of drive
        items: list["DriveItem"] = []
        drives: list["Drive"] = site.drives.get_all().execute_query()
        for drive in drives:
            items.extend(drive.root.children.get_all().execute_query())
        return items

    def map_permission(self, permission: "Permission") -> dict:
        return {
            "id": permission.id,
            "roles": list(permission.roles),
            "share_id": permission.share_id,
            "has_password": permission.has_password,
            "link": permission.link.to_json(),
            "granted_to_identities": permission.granted_to_identities.to_json(),
            "granted_to": permission.granted_to.to_json(),
            "granted_to_v2": permission.granted_to_v2.to_json(),
            "granted_to_identities_v2": permission.granted_to_identities_v2.to_json(),
            "invitation": permission.invitation.to_json(),
        }

    def enrich_permissions_on_files(self, all_file_data: list[FileData], site_url: str) -> None:
        logger.debug("Enriching permissions on files")
        permission_client = self.connection_config.get_permissions_client()
        if permission_client is None:
            return
        site = self.get_site(permissions_client=permission_client, site_url=site_url)
        existing_items = self.get_permissions_items(site=site)
        for file_data in all_file_data:
            etag = file_data.additional_metadata.get("ETag")
            if not etag:
                continue
            matching_items = list(filter(lambda x: x.etag == etag, existing_items))
            if not matching_items:
                continue
            if len(matching_items) > 1:
                logger.warning(
                    "Found multiple drive items with etag matching {}, skipping: {}".format(
                        etag, ", ".join([i.name for i in matching_items])
                    )
                )
                continue
            matching_item = matching_items[0]
            permissions: list["Permission"] = matching_item.permissions.get_all().execute_query()
            permissions_data = [
                self.map_permission(permission=permission) for permission in permissions
            ]
            file_data.metadata.permissions_data = permissions_data

    @property
    def process_permissions(self) -> bool:
        return (
            self.connection_config.permissions_config.permissions_tenant
            and self.connection_config.permissions_config.permissions_client_cred
            and self.connection_config.permissions_config.permissions_application_id
        )

    def run(self, **kwargs: Any) -> Generator[FileData, None, None]:
        client = self.connection_config.get_client()
        root_folder = self.get_root(client=client)
        logger.debug(f"processing content from path: {self.index_config.path}")
        if not self.index_config.omit_files:
            files = self.list_files(root_folder, recursive=self.index_config.recursive)
            file_data = [self.file_to_file_data(file=file, client=client) for file in files]
            if self.process_permissions:
                self.enrich_permissions_on_files(
                    all_file_data=file_data, site_url=self.get_site_url(client=client)
                )
            for file in file_data:
                yield file
        if not self.index_config.omit_pages:
            pages = self.list_pages(client=client)
            for page in pages:
                file_data = self.page_to_file_data(site_page=page)
                file_data.metadata.record_locator["site_url"] = client.base_url
                yield file_data


@dataclass
class SharepointDownloaderConfig(DownloaderConfig):
    pass


@dataclass
class SharepointDownloader(Downloader):
    connection_config: SharepointConnectionConfig
    download_config: SharepointDownloaderConfig
    connector_type: str = CONNECTOR_TYPE

    def get_download_path(self, file_data: FileData) -> Path:
        content_type = file_data.additional_metadata.get("sharepoint_content_type")
        rel_path = file_data.source_identifiers.fullpath
        rel_path = rel_path[1:] if rel_path.startswith("/") else rel_path
        download_path = self.download_dir / Path(rel_path)
        if content_type == SharepointContentType.SITEPAGE.value:
            # Update output extension to html if site page
            download_path = download_path.with_suffix(".html")
        return download_path

    def get_document(self, file_data: FileData) -> DownloadResponse:
        client: "ClientContext" = self.connection_config.get_client()
        file: "File" = client.web.get_file_by_id(unique_id=file_data.identifier)
        download_path = self.get_download_path(file_data=file_data)
        download_path.parent.mkdir(parents=True, exist_ok=True)
        logger.debug(
            f"writing document content {file_data.source_identifiers.fullpath} to {download_path}"
        )
        with download_path.open("wb") as f:
            file.download(f).execute_query()
        return self.generate_download_response(file_data=file_data, download_path=download_path)

    def get_site_page(self, file_data: FileData) -> DownloadResponse:
        # TODO fetch comments for site page as well
        from lxml import etree, html

        canvas_content_raw = file_data.additional_metadata.get("CanvasContent1")
        layout_web_parts_content_raw = file_data.additional_metadata.get("LayoutWebpartsContent")
        html_content = []
        if layout_web_parts_content_raw:
            layout_web_parts_content = json.loads(layout_web_parts_content_raw)
            for web_part in layout_web_parts_content:
                properties = web_part.get("properties", {})
                if title := properties.get("title"):
                    html_content.append(f"<title>{title}</title>")
        if canvas_content_raw:
            canvas_content = json.loads(canvas_content_raw)
            for content in canvas_content:
                if inner_html := content.get("innerHTML"):
                    html_content.append(inner_html)
        htmls = "".join(html_content)
        content = f"<div>{htmls}</div>"
        document = html.fromstring(content)
        download_path = self.get_download_path(file_data=file_data)
        download_path.parent.mkdir(parents=True, exist_ok=True)
        logger.debug(
            f"writing site page content {file_data.source_identifiers.filename} to {download_path}"
        )
        with download_path.open("w") as f:
            f.write(etree.tostring(document, encoding="unicode", pretty_print=True))
        return self.generate_download_response(file_data=file_data, download_path=download_path)

    def run(self, file_data: FileData, **kwargs: Any) -> download_responses:
        content_type = file_data.additional_metadata.get("sharepoint_content_type")
        if not content_type:
            raise ValueError(
                f"Missing sharepoint_content_type metadata: {file_data.additional_metadata}"
            )
        if content_type == SharepointContentType.DOCUMENT.value:
            return self.get_document(file_data=file_data)
        elif content_type == SharepointContentType.SITEPAGE.value:
            return self.get_site_page(file_data=file_data)


sharepoint_source_entry = SourceRegistryEntry(
    connection_config=SharepointConnectionConfig,
    indexer_config=SharepointIndexerConfig,
    indexer=SharepointIndexer,
    downloader_config=SharepointDownloaderConfig,
    downloader=SharepointDownloader,
)



import json
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional

from unstructured import __name__ as integration_name
from unstructured.__version__ import __version__ as integration_version

from unstructured_ingest.enhanced_dataclass import enhanced_field
from unstructured_ingest.utils.data_prep import batch_generator
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.interfaces import (
    AccessConfig,
    ConnectionConfig,
    FileData,
    UploadContent,
    Uploader,
    UploaderConfig,
    UploadStager,
    UploadStagerConfig,
)
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.processes.connector_registry import (
    DestinationRegistryEntry,
)

if TYPE_CHECKING:
    from astrapy.db import AstraDBCollection

CONNECTOR_TYPE = "astra"


@dataclass
class AstraAccessConfig(AccessConfig):
    token: str
    api_endpoint: str


@dataclass
class AstraConnectionConfig(ConnectionConfig):
    connection_type: str = CONNECTOR_TYPE
    access_config: AstraAccessConfig = enhanced_field(sensitive=True)


@dataclass
class AstraUploadStagerConfig(UploadStagerConfig):
    pass


@dataclass
class AstraUploadStager(UploadStager):
    upload_stager_config: AstraUploadStagerConfig = field(
        default_factory=lambda: AstraUploadStagerConfig()
    )

    def conform_dict(self, element_dict: dict) -> dict:
        return {
            "$vector": element_dict.pop("embeddings", None),
            "content": element_dict.pop("text", None),
            "metadata": element_dict,
        }

    def run(
        self,
        elements_filepath: Path,
        file_data: FileData,
        output_dir: Path,
        output_filename: str,
        **kwargs: Any,
    ) -> Path:
        with open(elements_filepath) as elements_file:
            elements_contents = json.load(elements_file)
        conformed_elements = []
        for element in elements_contents:
            conformed_elements.append(self.conform_dict(element_dict=element))
        output_path = Path(output_dir) / Path(f"{output_filename}.json")
        with open(output_path, "w") as output_file:
            json.dump(conformed_elements, output_file)
        return output_path


@dataclass
class AstraUploaderConfig(UploaderConfig):
    collection_name: str
    embedding_dimension: int
    namespace: Optional[str] = None
    requested_indexing_policy: Optional[dict[str, Any]] = None
    batch_size: int = 20


@dataclass
class AstraUploader(Uploader):
    connection_config: AstraConnectionConfig
    upload_config: AstraUploaderConfig
    connector_type: str = CONNECTOR_TYPE

    @requires_dependencies(["astrapy"], extras="astra")
    def get_collection(self) -> "AstraDBCollection":
        from astrapy.db import AstraDB

        # Get the collection_name and embedding dimension
        collection_name = self.upload_config.collection_name
        embedding_dimension = self.upload_config.embedding_dimension
        requested_indexing_policy = self.upload_config.requested_indexing_policy

        # If the user has requested an indexing policy, pass it to the AstraDB
        options = {"indexing": requested_indexing_policy} if requested_indexing_policy else None

        # Build the Astra DB object.
        # caller_name/version for AstraDB tracking
        astra_db = AstraDB(
            api_endpoint=self.connection_config.access_config.api_endpoint,
            token=self.connection_config.access_config.token,
            namespace=self.upload_config.namespace,
            caller_name=integration_name,
            caller_version=integration_version,
        )

        # Create and connect to the newly created collection
        astra_db_collection = astra_db.create_collection(
            collection_name=collection_name,
            dimension=embedding_dimension,
            options=options,
        )
        return astra_db_collection

    def run(self, contents: list[UploadContent], **kwargs: Any) -> None:
        elements_dict = []
        for content in contents:
            with open(content.path) as elements_file:
                elements = json.load(elements_file)
                elements_dict.extend(elements)

        logger.info(
            f"writing {len(elements_dict)} objects to destination "
            f"collection {self.upload_config.collection_name}"
        )

        astra_batch_size = self.upload_config.batch_size
        collection = self.get_collection()

        for chunk in batch_generator(elements_dict, astra_batch_size):
            collection.insert_many(chunk)


astra_destination_entry = DestinationRegistryEntry(
    connection_config=AstraConnectionConfig,
    upload_stager_config=AstraUploadStagerConfig,
    upload_stager=AstraUploadStager,
    uploader_config=AstraUploaderConfig,
    uploader=AstraUploader,
)



import asyncio
import hashlib
import json
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Optional, TypedDict

from unstructured.staging.base import elements_to_dicts

from unstructured_ingest.v2.interfaces import FileData
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.pipeline.interfaces import PipelineStep
from unstructured_ingest.v2.pipeline.utils import sterilize_dict
from unstructured_ingest.v2.processes.chunker import Chunker

STEP_ID = "chunk"


class ChunkStepResponse(TypedDict):
    file_data_path: str
    path: str


@dataclass
class ChunkStep(PipelineStep):
    process: Chunker
    identifier: str = STEP_ID

    def __str__(self):
        return f"{self.identifier} ({self.process.config.chunking_strategy})"

    def __post_init__(self):
        config = (
            sterilize_dict(self.process.config.to_dict(redact_sensitive=True))
            if self.process.config
            else None
        )
        logger.info(f"Created {self.identifier} with configs: {config}")

    def should_chunk(self, filepath: Path, file_data: FileData) -> bool:
        if self.context.reprocess or file_data.reprocess:
            return True
        return not filepath.exists()

    def get_output_filepath(self, filename: Path) -> Path:
        hashed_output_file = f"{self.get_hash(extras=[filename.name])}.json"
        filepath = (self.cache_dir / hashed_output_file).resolve()
        filepath.parent.mkdir(parents=True, exist_ok=True)
        return filepath

    def _save_output(self, output_filepath: str, chunked_content: list[dict]):
        with open(str(output_filepath), "w") as f:
            logger.debug(f"Writing chunker output to: {output_filepath}")
            json.dump(chunked_content, f, indent=2)

    async def _run_async(
        self, fn: Callable, path: str, file_data_path: str, **kwargs
    ) -> ChunkStepResponse:
        path = Path(path)
        file_data = FileData.from_file(path=file_data_path)
        output_filepath = self.get_output_filepath(filename=path)
        if not self.should_chunk(filepath=output_filepath, file_data=file_data):
            logger.debug(f"Skipping chunking, output already exists: {output_filepath}")
            return ChunkStepResponse(file_data_path=file_data_path, path=str(output_filepath))
        fn_kwargs = {"elements_filepath": path}
        if not asyncio.iscoroutinefunction(fn):
            chunked_content_raw = fn(**fn_kwargs)
        elif semaphore := self.context.semaphore:
            async with semaphore:
                chunked_content_raw = await fn(**fn_kwargs)
        else:
            chunked_content_raw = await fn(**fn_kwargs)
        self._save_output(
            output_filepath=str(output_filepath),
            chunked_content=elements_to_dicts(chunked_content_raw),
        )
        return ChunkStepResponse(file_data_path=file_data_path, path=str(output_filepath))

    def get_hash(self, extras: Optional[list[str]]) -> str:
        hashable_string = json.dumps(
            self.process.config.to_dict(), sort_keys=True, ensure_ascii=True
        )
        if extras:
            hashable_string += "".join(extras)
        return hashlib.sha256(hashable_string.encode()).hexdigest()[:12]



from dataclasses import dataclass

import click

from unstructured_ingest.v2.cli.base import DestCmd
from unstructured_ingest.v2.cli.interfaces import CliConfig
from unstructured_ingest.v2.processes.connectors.pinecone import CONNECTOR_TYPE


@dataclass
class PineconeCliConnectionConfig(CliConfig):
    @staticmethod
    def get_cli_options() -> list[click.Option]:
        options = [
            click.Option(
                ["--api-key"],
                required=True,
                type=str,
                help="API key for Pinecone.",
            ),
            click.Option(
                ["--index-name"],
                required=True,
                type=str,
                help="Name of the index to connect to. Example: my-index",
            ),
            click.Option(
                ["--environment"],
                required=True,
                type=str,
                help="Environment to connect to. Example: us-east-1",
            ),
        ]
        return options


@dataclass
class PineconeCliUploaderConfig(CliConfig):
    @staticmethod
    def get_cli_options() -> list[click.Option]:
        options = [
            click.Option(
                ["--batch-size"],
                default=100,
                type=int,
                help="Number of records per batch",
            ),
            click.Option(
                ["--num-processes"],
                default=4,
                type=int,
                help="Number of processes to use for uploading",
            ),
        ]
        return options


pinecone_dest_cmd = DestCmd(
    cmd_name=CONNECTOR_TYPE,
    connection_config=PineconeCliConnectionConfig,
    uploader_config=PineconeCliUploaderConfig,
)



from dataclasses import dataclass

import click

from unstructured_ingest.v2.cli.base import DestCmd, SrcCmd
from unstructured_ingest.v2.cli.cmds.fsspec.fsspec import (
    FsspecCliDownloadConfig,
    FsspecCliIndexerConfig,
    FsspecCliUploaderConfig,
)
from unstructured_ingest.v2.cli.interfaces import CliConfig
from unstructured_ingest.v2.processes.connectors.fsspec.azure import (
    CONNECTOR_TYPE,
)


@dataclass
class AzureCliDownloadConfig(FsspecCliDownloadConfig):
    pass


@dataclass
class AzureCliIndexerConfig(FsspecCliIndexerConfig):
    pass


@dataclass
class AzureCliConnectionConfig(CliConfig):
    @staticmethod
    def get_cli_options() -> list[click.Option]:
        options = [
            click.Option(
                ["--account-key"],
                default=None,
                help="The storage account key. This is used for shared key "
                "authentication. If any of account key, sas token or "
                "client_id are not specified, anonymous access will be used.",
            ),
            click.Option(
                ["--account-name"],
                default=None,
                help="The storage account name. This is used to authenticate "
                "requests signed with an account key and to construct "
                "the storage endpoint. It is required unless a connection "
                "string is given, or if a custom domain is used with "
                "anonymous authentication.",
            ),
            click.Option(
                ["--connection-string"],
                default=None,
                help="If specified, this will override all other parameters. See "
                "http://azure.microsoft.com/en-us/documentation/articles/storage-configure-connection-string/ "  # noqa: E501
                "for the connection string format.",
            ),
            click.Option(
                ["--sas_token"],
                default=None,
                help="A shared access signature token to use to authenticate "
                "requests instead of the account key. If account key and "
                "sas token are both specified, account key will be used "
                "to sign. If any of account key, sas token or client_id "
                "are not specified, anonymous access will be used.",
            ),
        ]
        return options


@dataclass
class AzureUploaderConfig(FsspecCliUploaderConfig):
    pass


azure_src_cmd = SrcCmd(
    cmd_name=CONNECTOR_TYPE,
    indexer_config=AzureCliIndexerConfig,
    connection_config=AzureCliConnectionConfig,
    downloader_config=AzureCliDownloadConfig,
)

azure_dest_cmd = DestCmd(
    cmd_name=CONNECTOR_TYPE,
    connection_config=AzureCliConnectionConfig,
    uploader_config=AzureUploaderConfig,
)



from dataclasses import dataclass

import click

from unstructured_ingest.v2.cli.base import DestCmd
from unstructured_ingest.v2.cli.interfaces import CliConfig
from unstructured_ingest.v2.processes.connectors.singlestore import CONNECTOR_TYPE


@dataclass
class SingleStoreCliConnectionConfig(CliConfig):
    @staticmethod
    def get_cli_options() -> list[click.Option]:
        options = [
            click.Option(
                ["--host"],
                required=False,
                type=str,
                default=None,
                help="SingleStore host",
            ),
            click.Option(
                ["--port"],
                required=False,
                type=int,
                default=None,
                help="SingleStore port",
            ),
            click.Option(
                ["--user"],
                required=False,
                type=str,
                default=None,
                help="SingleStore user",
            ),
            click.Option(
                ["--password"],
                required=False,
                type=str,
                default=None,
                help="SingleStore password",
            ),
            click.Option(
                ["--database"],
                required=False,
                type=str,
                default=None,
                help="SingleStore database",
            ),
        ]
        return options


@dataclass
class SingleStoreCliUploaderConfig(CliConfig):
    @staticmethod
    def get_cli_options() -> list[click.Option]:
        options = [
            click.Option(
                ["--drop-empty-cols"],
                required=False,
                type=bool,
                is_flag=True,
                default=False,
                help="Drop any columns that have no data",
            ),
        ]
        return options


@dataclass
class SingleStoreCliUploadStagerConfig(CliConfig):
    @staticmethod
    def get_cli_options() -> list[click.Option]:
        return [
            click.Option(
                ["--table-name"],
                required=False,
                type=str,
                help="SingleStore table to write contents to",
            ),
            click.Option(
                ["--batch-size"],
                required=False,
                type=click.IntRange(min=1),
                help="Batch size when writing to SingleStore",
            ),
        ]


singlestore_dest_cmd = DestCmd(
    cmd_name=CONNECTOR_TYPE,
    connection_config=SingleStoreCliConnectionConfig,
    uploader_config=SingleStoreCliUploaderConfig,
    upload_stager_config=SingleStoreCliUploadStagerConfig,
)



import click

from unstructured_ingest.cli import dest, src
from unstructured_ingest.v2.cli.cmds import dest as dest_v2
from unstructured_ingest.v2.cli.cmds import src as src_v2


@click.group()
def ingest():
    pass


def get_cmd() -> click.Command:
    """Construct and return a Click command object representing the main command for the CLI.

    This function adds all dest_subcommand(s) to each src_subcommand, and adds all of those
    to the main command as nested subcommands.
    """
    cmd = ingest
    src_dict = {s.name: s for s in src}
    dest_dict = {d.name: d for d in dest}
    for s in src_v2:
        src_dict[s.name] = s
    for d in dest_v2:
        dest_dict[d.name] = d
    # Add all subcommands
    for src_subcommand in src_dict.values():
        # Add all destination subcommands
        for dest_subcommand in dest_dict.values():
            src_subcommand.add_command(dest_subcommand)
        cmd.add_command(src_subcommand)
    return cmd



import json
import os
import typing as t
from dataclasses import dataclass
from html import unescape
from pathlib import Path
from urllib.parse import urlparse

from unstructured_ingest.enhanced_dataclass import enhanced_field
from unstructured_ingest.error import SourceConnectionError, SourceConnectionNetworkError
from unstructured_ingest.interfaces import (
    AccessConfig,
    BaseConnectorConfig,
    BaseSingleIngestDoc,
    BaseSourceConnector,
    IngestDocCleanupMixin,
    SourceConnectorCleanupMixin,
    SourceMetadata,
)
from unstructured_ingest.interfaces import PermissionsConfig as SharepointPermissionsConfig
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.utils.string_and_date_utils import ensure_isoformat_datetime

if t.TYPE_CHECKING:
    from office365.sharepoint.client_context import ClientContext
    from office365.sharepoint.files.file import File
    from office365.sharepoint.publishing.pages.page import SitePage

MAX_MB_SIZE = 512_000_000
CONTENT_LABELS = ["CanvasContent1", "LayoutWebpartsContent1", "TimeCreated"]


@dataclass
class SharepointAccessConfig(AccessConfig):
    client_cred: str = enhanced_field(repr=False, sensitive=True)


@dataclass
class SimpleSharepointConfig(BaseConnectorConfig):
    access_config: SharepointAccessConfig
    client_id: str
    site: str
    path: str
    process_pages: bool = enhanced_field(default=True, init=False)
    recursive: bool = False
    files_only: bool = False
    permissions_config: t.Optional[SharepointPermissionsConfig] = None

    def __post_init__(self):
        if not (self.client_id and self.access_config.client_cred and self.site):
            raise ValueError(
                "Please provide one of the following mandatory values:"
                "\n--client-id\n--client-cred\n--site",
            )
        self.process_pages = not self.files_only

    @requires_dependencies(["office365"], extras="sharepoint")
    def get_site_client(self, site_url: str = "") -> "ClientContext":
        from office365.runtime.auth.client_credential import ClientCredential
        from office365.sharepoint.client_context import ClientContext

        try:
            site_client = ClientContext(site_url or self.site).with_credentials(
                ClientCredential(self.client_id, self.access_config.client_cred),
            )
        except Exception:
            logger.error("Couldn't set Sharepoint client.")
            raise
        return site_client

    def get_permissions_client(self):
        try:
            permissions_connector = SharepointPermissionsConnector(self.permissions_config)
            assert permissions_connector.access_token
            return permissions_connector
        except Exception as e:
            logger.error("Couldn't obtain Sharepoint permissions ingestion access token:", e)


@dataclass
class SharepointIngestDoc(IngestDocCleanupMixin, BaseSingleIngestDoc):
    connector_config: SimpleSharepointConfig
    site_url: str
    server_path: str
    is_page: bool
    file_path: str
    registry_name: str = "sharepoint"

    def __post_init__(self):
        self.extension = Path(self.file_path).suffix if not self.is_page else ".html"
        self.extension = ".html" if self.extension == ".aspx" else self.extension
        if not self.extension:
            raise ValueError("Unsupported file without extension.")

        self._set_download_paths()

    def _set_download_paths(self) -> None:
        """Parses the folder structure from the source and creates the download and output paths"""
        download_path = Path(f"{self.read_config.download_dir}")
        output_path = Path(f"{self.processor_config.output_dir}")
        parent = Path(self.file_path).with_suffix(self.extension)
        self.download_dir = (download_path / parent.parent).resolve()
        self.download_filepath = (download_path / parent).resolve()
        output_filename = str(parent) + ".json"
        self.output_dir = (output_path / parent.parent).resolve()
        self.output_filepath = (output_path / output_filename).resolve()

    @property
    def filename(self):
        return Path(self.download_filepath).resolve()

    @property
    def _output_filename(self):
        return Path(self.output_filepath).resolve()

    @property
    def record_locator(self) -> t.Optional[t.Dict[str, t.Any]]:
        return {
            "server_path": self.server_path,
            "site_url": self.site_url,
        }

    @SourceConnectionNetworkError.wrap
    @requires_dependencies(["office365"], extras="sharepoint")
    def _fetch_file(self, properties_only: bool = False):
        """Retrieves the actual page/file from the Sharepoint instance"""
        from office365.runtime.client_request_exception import ClientRequestException

        site_client = self.connector_config.get_site_client(self.site_url)

        try:
            if self.is_page:
                file = site_client.web.get_file_by_server_relative_path("/" + self.server_path)
                file = file.listItemAllFields.select(CONTENT_LABELS).get().execute_query()
            else:
                file = site_client.web.get_file_by_server_relative_url(self.server_path)
                if properties_only:
                    file = file.get().execute_query()
        except ClientRequestException as e:
            if e.response.status_code == 404:
                return None
            raise
        return file

    def _fetch_page(self):
        site_client = self.connector_config.get_site_client(self.site_url)
        try:
            page = (
                site_client.site_pages.pages.get_by_url(self.server_path)
                .expand(["FirstPublished", "Modified", "Version"])
                .get()
                .execute_query()
            )
        except Exception as e:
            logger.error(f"Failed to retrieve page {self.server_path} from site {self.site_url}")
            logger.error(e)
            return None
        return page

    def update_permissions_data(self):
        def parent_name_matches(parent_type, permissions_filename, ingest_doc_filepath):
            permissions_filename = permissions_filename.split("_SEP_")
            ingest_doc_filepath = ingest_doc_filepath.split("/")

            if parent_type == "sites":
                return permissions_filename[0] == ingest_doc_filepath[1]

            elif parent_type == "SitePages" or parent_type == "Shared Documents":
                return True

        permissions_data = None
        permissions_dir = Path(self.processor_config.output_dir) / "permissions_data"

        if permissions_dir.is_dir():
            parent_type = self.file_path.split("/")[0]

            if parent_type == "sites":
                read_dir = permissions_dir / "sites"
            elif parent_type == "SitePages" or parent_type == "Shared Documents":
                read_dir = permissions_dir / "other"
            else:
                read_dir = permissions_dir / "other"

            for filename in os.listdir(read_dir):
                permissions_docname = os.path.splitext(filename)[0].split("_SEP_")[1]
                ingestdoc_docname = self.file_path.split("/")[-1]

                if ingestdoc_docname == permissions_docname and parent_name_matches(
                    parent_type=parent_type,
                    permissions_filename=filename,
                    ingest_doc_filepath=self.file_path,
                ):
                    with open(read_dir / filename) as f:
                        permissions_data = json.loads(f.read())

        return permissions_data

    def update_source_metadata(self, **kwargs):
        if self.is_page:
            page = self._fetch_page()
            if page is None:
                self.source_metadata = SourceMetadata(
                    exists=False,
                )
                return
            self.source_metadata = SourceMetadata(
                date_created=page.get_property("FirstPublished", None),
                date_modified=page.get_property("Modified", None),
                version=page.get_property("Version", ""),
                source_url=page.absolute_url,
                exists=True,
                permissions_data=(
                    self.update_permissions_data()
                    if self.connector_config.permissions_config
                    else None
                ),
            )
            return

        file = self._fetch_file(True)
        if file is None:
            self.source_metadata = SourceMetadata(
                exists=False,
            )
            return
        self.source_metadata = SourceMetadata(
            date_created=ensure_isoformat_datetime(timestamp=file.time_created),
            date_modified=ensure_isoformat_datetime(timestamp=file.time_last_modified),
            version=file.major_version,
            source_url=file.properties.get("LinkingUrl", None),
            exists=True,
            permissions_data=(
                self.update_permissions_data() if self.connector_config.permissions_config else None
            ),
        )

    def _download_page(self):
        """Formats and saves locally page content"""
        content = self._fetch_file()
        self.update_source_metadata()
        pld = (content.properties.get("LayoutWebpartsContent1", "") or "") + (
            content.properties.get("CanvasContent1", "") or ""
        )
        if pld != "":
            pld = unescape(pld)
        else:
            logger.info(
                f"Page {self.server_path} has no retrievable content. \
                    Dumping empty doc.",
            )
            pld = "<div></div>"

        self.output_dir.mkdir(parents=True, exist_ok=True)
        if not self.download_dir.is_dir():
            logger.debug(f"Creating directory: {self.download_dir}")
            self.download_dir.mkdir(parents=True, exist_ok=True)
        with self.filename.open(mode="w") as f:
            f.write(pld)
        logger.info(f"File downloaded: {self.filename}")

    def _download_file(self):
        file = self._fetch_file()
        self.update_source_metadata()
        fsize = file.length
        self.output_dir.mkdir(parents=True, exist_ok=True)

        if not self.download_dir.is_dir():
            logger.debug(f"Creating directory: {self.download_dir}")
            self.download_dir.mkdir(parents=True, exist_ok=True)

        if fsize > MAX_MB_SIZE:
            logger.info(f"Downloading file with size: {fsize} bytes in chunks")
            with self.filename.open(mode="wb") as f:
                file.download_session(f, chunk_size=1024 * 1024 * 100).execute_query()
        else:
            with self.filename.open(mode="wb") as f:
                file.download(f).execute_query()
        logger.info(f"File downloaded: {self.filename}")

    @BaseSingleIngestDoc.skip_if_file_exists
    @SourceConnectionError.wrap
    @requires_dependencies(["office365"])
    def get_file(self):
        if self.is_page:
            self._download_page()
        else:
            self._download_file()
        return


@dataclass
class SharepointSourceConnector(SourceConnectorCleanupMixin, BaseSourceConnector):
    connector_config: SimpleSharepointConfig

    def check_connection(self):
        try:
            site_client = self.connector_config.get_site_client()
            site_client.site_pages.pages.get().execute_query()
        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise SourceConnectionError(f"failed to validate connection: {e}")

    @requires_dependencies(["office365"], extras="sharepoint")
    def _list_files(self, folder, recursive) -> t.List["File"]:
        from office365.runtime.client_request_exception import ClientRequestException

        try:
            objects = folder.expand(["Files", "Folders"]).get().execute_query()
            files = list(objects.files)
            if not recursive:
                return files
            for f in objects.folders:
                if "/Forms" in f.serverRelativeUrl:
                    continue
                files += self._list_files(f, recursive)
            return files
        except ClientRequestException as e:
            if e.response.status_code != 404:
                logger.info("Caught an error while processing documents %s", e.response.text)
            return []

    def _prepare_ingest_doc(self, obj: t.Union["File", "SitePage"], base_url, is_page=False):
        if is_page:
            file_path = obj.get_property("Url", "")
            server_path = file_path if file_path[0] != "/" else file_path[1:]
            if (url_path := (urlparse(base_url).path)) and (url_path != "/"):
                file_path = url_path[1:] + "/" + file_path
        else:
            server_path = obj.serverRelativeUrl
            file_path = obj.serverRelativeUrl[1:]

        return SharepointIngestDoc(
            processor_config=self.processor_config,
            read_config=self.read_config,
            connector_config=self.connector_config,
            site_url=base_url,
            server_path=server_path,
            is_page=is_page,
            file_path=file_path,
        )

    @requires_dependencies(["office365"], extras="sharepoint")
    def _list_pages(self, site_client) -> list:
        from office365.runtime.client_request_exception import ClientRequestException

        try:
            site_pages = site_client.site_pages.pages.get().execute_query()
        except ClientRequestException as e:
            logger.info(
                "Caught an error while retrieving site pages from %s \n%s",
                site_client.base_url,
                e.response.text,
            )
            return []

        return [self._prepare_ingest_doc(page, site_client.base_url, True) for page in site_pages]

    def _ingest_site_docs(self, site_client) -> t.List["SharepointIngestDoc"]:
        root_folder = site_client.web.get_folder_by_server_relative_path(self.connector_config.path)
        files = self._list_files(root_folder, self.connector_config.recursive)
        if not files:
            logger.info(
                f"No processable files at path {self.connector_config.path}\
                for site {site_client.base_url}",
            )
        output = []
        for file in files:
            try:
                output.append(self._prepare_ingest_doc(file, site_client.base_url))
            except ValueError as e:
                logger.error("Unable to process file %s", file.properties["Name"])
                logger.error(e)
        if self.connector_config.process_pages:
            page_output = self._list_pages(site_client)
            if not page_output:
                logger.info(f"Couldn't process pages for site {site_client.base_url}")
            output = output + page_output
        return output

    def initialize(self):
        pass

    def get_ingest_docs(self):
        base_site_client = self.connector_config.get_site_client()

        if not all(
            getattr(self.connector_config.permissions_config, attr, False)
            for attr in ["application_id", "client_cred", "tenant"]
        ):
            logger.info(
                "Permissions config is not fed with 'application_id', 'client_cred' and 'tenant'."
                "Skipping permissions ingestion.",
            )
        else:
            permissions_client = self.connector_config.get_permissions_client()
            if permissions_client:
                permissions_client.write_all_permissions(self.processor_config.output_dir)

        if not base_site_client.is_tenant:
            return self._ingest_site_docs(base_site_client)
        tenant = base_site_client.tenant
        tenant_sites = tenant.get_site_properties_from_sharepoint_by_filters().execute_query()
        tenant_sites = {s.url for s in tenant_sites if (s.url is not None)}
        ingest_docs: t.List[SharepointIngestDoc] = []
        for site_url in tenant_sites:
            logger.info(f"Processing docs for site: {site_url}")
            site_client = self.connector_config.get_site_client(site_url)
            ingest_docs = ingest_docs + self._ingest_site_docs(site_client)
        return ingest_docs


@dataclass
class SharepointPermissionsConnector:
    def __init__(self, permissions_config):
        self.permissions_config: SharepointPermissionsConfig = permissions_config
        self.initialize()

    def initialize(self):
        self.access_token: str = self.get_access_token()

    @requires_dependencies(["requests"], extras="sharepoint")
    def get_access_token(self) -> str:
        import requests

        url = (
            f"https://login.microsoftonline.com/{self.permissions_config.tenant}/oauth2/v2.0/token"
        )
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        data = {
            "client_id": self.permissions_config.application_id,
            "scope": "https://graph.microsoft.com/.default",
            "client_secret": self.permissions_config.client_cred,
            "grant_type": "client_credentials",
        }
        response = requests.post(url, headers=headers, data=data)
        return response.json()["access_token"]

    def validated_response(self, response):
        if response.status_code == 200:
            return response.json()
        else:
            logger.info(f"Request failed with status code {response.status_code}:")
            logger.info(response.text)

    @requires_dependencies(["requests"], extras="sharepoint")
    def get_sites(self):
        import requests

        url = "https://graph.microsoft.com/v1.0/sites"
        params = {
            "$select": "webUrl, id",
        }

        headers = {
            "Authorization": f"Bearer {self.access_token}",
        }

        response = requests.get(url, params=params, headers=headers)
        return self.validated_response(response)

    @requires_dependencies(["requests"], extras="sharepoint")
    def get_drives(self, site):
        import requests

        url = f"https://graph.microsoft.com/v1.0/sites/{site}/drives"

        headers = {
            "Authorization": f"Bearer {self.access_token}",
        }

        response = requests.get(url, headers=headers)

        return self.validated_response(response)

    @requires_dependencies(["requests"], extras="sharepoint")
    def get_drive_items(self, site, drive_id):
        import requests

        url = f"https://graph.microsoft.com/v1.0/sites/{site}/drives/{drive_id}/root/children"

        headers = {
            "Authorization": f"Bearer {self.access_token}",
        }

        response = requests.get(url, headers=headers)

        return self.validated_response(response)

    def extract_site_name_from_weburl(self, weburl):
        split_path = urlparse(weburl).path.lstrip("/").split("/")

        if split_path[0] == "sites":
            return "sites", split_path[1]

        elif split_path[0] == "Shared%20Documents":
            return "Shared Documents", "Shared Documents"

        elif split_path[0] == "personal":
            return "Personal", "Personal"

        elif split_path[0] == "_layouts":
            return "layouts", "layouts"

        # if other weburl structures are found, additional logic might need to be implemented

        logger.warning(
            """Couldn't extract sitename, unknown site or parent type. Skipping permissions
            ingestion for the document with the URL:""",
            weburl,
        )

        return None, None

    @requires_dependencies(["requests"], extras="sharepoint")
    def get_permissions_for_drive_item(self, site, drive_id, item_id):
        import requests

        url = f"https://graph.microsoft.com/v1.0/sites/ \
        {site}/drives/{drive_id}/items/{item_id}/permissions"

        headers = {
            "Authorization": f"Bearer {self.access_token}",
        }

        response = requests.get(url, headers=headers)

        return self.validated_response(response)

    def write_all_permissions(self, output_dir):
        sites = [(site["id"], site["webUrl"]) for site in self.get_sites()["value"]]
        drive_ids = []

        logger.info("Obtaining drive data for sites for permissions (rbac)")
        for site_id, site_url in sites:
            drives = self.get_drives(site_id)
            if drives:
                drives_for_site = drives["value"]
                drive_ids.extend([(site_id, drive["id"]) for drive in drives_for_site])

        logger.info("Obtaining item data from drives for permissions (rbac)")
        item_ids = []
        for site, drive_id in drive_ids:
            drive_items = self.get_drive_items(site, drive_id)
            if drive_items:
                item_ids.extend(
                    [
                        (site, drive_id, item["id"], item["name"], item["webUrl"])
                        for item in drive_items["value"]
                    ],
                )

        permissions_dir = Path(output_dir) / "permissions_data"

        logger.info("Writing permissions data to disk")
        for site, drive_id, item_id, item_name, item_web_url in item_ids:
            res = self.get_permissions_for_drive_item(site, drive_id, item_id)
            if res:
                parent_type, parent_name = self.extract_site_name_from_weburl(item_web_url)

                if parent_type == "sites":
                    write_path = permissions_dir / "sites" / f"{parent_name}_SEP_{item_name}.json"

                elif parent_type == "Personal" or parent_type == "Shared Documents":
                    write_path = permissions_dir / "other" / f"{parent_name}_SEP_{item_name}.json"
                else:
                    write_path = permissions_dir / "other" / f"{parent_name}_SEP_{item_name}.json"

                if not Path(os.path.dirname(write_path)).is_dir():
                    os.makedirs(os.path.dirname(write_path))

                with open(write_path, "w") as f:
                    json.dump(res["value"], f)



import fnmatch
import glob
import os
import typing as t
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path

from unstructured_ingest.interfaces import (
    BaseConnectorConfig,
    BaseSingleIngestDoc,
    BaseSourceConnector,
    SourceMetadata,
)
from unstructured_ingest.logger import logger


@dataclass
class SimpleLocalConfig(BaseConnectorConfig):
    # Local specific options
    input_path: str
    recursive: bool = False
    file_glob: t.Optional[t.List[str]] = None

    def __post_init__(self):
        if os.path.isfile(self.input_path):
            self.input_path_is_file = True
        else:
            self.input_path_is_file = False


@dataclass
class LocalIngestDoc(BaseSingleIngestDoc):
    """Class encapsulating fetching a doc and writing processed results (but not
    doing the processing!).
    """

    connector_config: SimpleLocalConfig
    path: str
    registry_name: str = "local"

    @property
    def base_filename(self) -> t.Optional[str]:
        download_path = Path(self.connector_config.input_path).resolve()
        full_path = Path(self.filename).resolve()
        if download_path.is_file():
            download_path = download_path.parent
        relative_path = full_path.relative_to(download_path)
        return str(relative_path)

    @property
    def filename(self):
        """The filename of the local file to be processed"""
        return Path(self.path)

    def cleanup_file(self):
        """Not applicable to local file system"""

    def get_file(self):
        """Not applicable to local file system"""

    def update_source_metadata(self, **kwargs) -> None:
        try:
            out = os.lstat(self.path)
            self._source_metadata = SourceMetadata(
                exists=True,
                date_created=str(datetime.fromtimestamp(out.st_ctime)),
                date_modified=str(datetime.fromtimestamp(out.st_mtime)),
                permissions_data=[{"mode": out.st_mode}],
                source_url=self.path,
            )
        except FileNotFoundError:
            self._source_metadata = SourceMetadata(exists=False)

    @property
    def _output_filename(self) -> Path:
        """Returns output filename for the doc
        If input path argument is a file itself, it returns the filename of the doc.
        If input path argument is a folder, it returns the relative path of the doc.
        """
        input_path = Path(self.connector_config.input_path)
        basename = (
            f"{self.base_filename}.json"
            if input_path.is_file()
            else f"{Path(self.path).relative_to(input_path)}.json"
        )
        return Path(self.processor_config.output_dir) / basename


@dataclass
class LocalSourceConnector(BaseSourceConnector):
    """Objects of this class support fetching document(s) from local file system"""

    def check_connection(self):
        pass

    connector_config: SimpleLocalConfig

    def __post_init__(self):
        self.ingest_doc_cls: t.Type[LocalIngestDoc] = LocalIngestDoc

    def cleanup(self, cur_dir=None):
        """Not applicable to local file system"""

    def initialize(self):
        """Not applicable to local file system"""

    def _list_files(self):
        if self.connector_config.input_path_is_file:
            return glob.glob(f"{self.connector_config.input_path}")
        elif self.connector_config.recursive:
            return glob.glob(
                f"{self.connector_config.input_path}/**",
                recursive=self.connector_config.recursive,
            )
        else:
            return glob.glob(f"{self.connector_config.input_path}/*")

    def does_path_match_glob(self, path: str) -> bool:
        if self.connector_config.file_glob is None:
            return True
        patterns = self.connector_config.file_glob
        for pattern in patterns:
            if fnmatch.filter([path], pattern):
                return True
        logger.debug(f"The file {path!r} is discarded as it does not match any given glob.")
        return False

    def get_ingest_docs(self):
        return [
            self.ingest_doc_cls(
                connector_config=self.connector_config,
                processor_config=self.processor_config,
                read_config=self.read_config,
                path=file,
            )
            for file in self._list_files()
            if os.path.isfile(file) and self.does_path_match_glob(file)
        ]



from dataclasses import dataclass

import click

from unstructured_ingest.v2.cli.base import DestCmd
from unstructured_ingest.v2.cli.interfaces import CliConfig
from unstructured_ingest.v2.processes.connectors.sql import CONNECTOR_TYPE

SQL_DRIVERS = {"postgresql", "sqlite"}


@dataclass
class SQLCliConnectionConfig(CliConfig):
    @staticmethod
    def get_cli_options() -> list[click.Option]:
        options = [
            click.Option(
                ["--db-type"],
                required=True,
                type=click.Choice(SQL_DRIVERS),
                help="Type of the database backend",
            ),
            click.Option(
                ["--username"],
                default=None,
                type=str,
                help="DB username",
            ),
            click.Option(
                ["--password"],
                default=None,
                type=str,
                help="DB password",
            ),
            click.Option(
                ["--host"],
                default=None,
                type=str,
                help="DB host",
            ),
            click.Option(
                ["--port"],
                default=None,
                type=int,
                help="DB host connection port",
            ),
            click.Option(
                ["--database"],
                default=None,
                type=str,
                help="Database name. For sqlite databases, this is the path to the .db file.",
            ),
        ]
        return options


@dataclass
class SQLCliUploaderConfig(CliConfig):
    @staticmethod
    def get_cli_options() -> list[click.Option]:
        options = [
            click.Option(
                ["--batch-size"],
                default=100,
                type=int,
                help="Number of records per batch",
            )
        ]
        return options


@dataclass
class SQLCliUploadStagerConfig(CliConfig):
    @staticmethod
    def get_cli_options() -> list[click.Option]:
        return []


sql_dest_cmd = DestCmd(
    cmd_name=CONNECTOR_TYPE,
    connection_config=SQLCliConnectionConfig,
    uploader_config=SQLCliUploaderConfig,
    upload_stager_config=SQLCliUploadStagerConfig,
)



"""
Salesforce Connector
Able to download Account, Case, Campaign, EmailMessage, Lead
Salesforce returns everything as a list of json.
This saves each entry as a separate file to be partitioned.
Using JWT authorization
https://developer.salesforce.com/docs/atlas.en-us.sfdx_dev.meta/sfdx_dev/sfdx_dev_auth_key_and_cert.htm
https://developer.salesforce.com/docs/atlas.en-us.sfdx_dev.meta/sfdx_dev/sfdx_dev_auth_connected_app.htm
"""

import json
from collections import OrderedDict
from dataclasses import dataclass, field
from email.utils import formatdate
from pathlib import Path
from string import Template
from textwrap import dedent
from typing import TYPE_CHECKING, Any, Generator, Type

from dateutil import parser
from unstructured.documents.elements import DataSourceMetadata

from unstructured_ingest.enhanced_dataclass import enhanced_field
from unstructured_ingest.error import SourceConnectionNetworkError
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Downloader,
    DownloaderConfig,
    DownloadResponse,
    FileData,
    Indexer,
    IndexerConfig,
    SourceIdentifiers,
)
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.processes.connector_registry import (
    SourceRegistryEntry,
)


class MissingCategoryError(Exception):
    """There are no categories with that name."""


CONNECTOR_TYPE = "salesforce"

if TYPE_CHECKING:
    from simple_salesforce import Salesforce

SALESFORCE_API_VERSION = "57.0"

# TODO: Add more categories as needed
ACCEPTED_CATEGORIES: list[str] = ["Account", "Case", "Campaign", "EmailMessage", "Lead"]

# Generic minimal email template used only
# to process EmailMessage records as .eml files
EMAIL_TEMPLATE = Template(
    """MIME-Version: 1.0
Date: $date
Message-ID: $message_identifier
Subject: $subject
From: $from_email
To: $to_email
Content-Type: multipart/alternative; boundary="00000000000095c9b205eff92630"
--00000000000095c9b205eff92630
Content-Type: text/plain; charset="UTF-8"
$textbody
--00000000000095c9b205eff92630
Content-Type: text/html; charset="UTF-8"
$htmlbody
--00000000000095c9b205eff92630--
""",
)


@dataclass
class SalesforceAccessConfig(AccessConfig):
    consumer_key: str
    private_key: str

    @requires_dependencies(["cryptography"])
    def get_private_key_value_and_type(self) -> tuple[str, Type]:
        from cryptography.hazmat.primitives import serialization

        try:
            serialization.load_pem_private_key(data=self.private_key.encode("utf-8"), password=None)
        except ValueError:
            pass
        else:
            return self.private_key, str

        if Path(self.private_key).is_file():
            return self.private_key, Path

        raise ValueError("private_key does not contain PEM private key or path")


@dataclass
class SalesforceConnectionConfig(ConnectionConfig):
    username: str
    access_config: SalesforceAccessConfig = enhanced_field(sensitive=True)

    @requires_dependencies(["simple_salesforce"], extras="salesforce")
    def get_client(self) -> "Salesforce":
        from simple_salesforce import Salesforce

        pkey_value, pkey_type = self.access_config.get_private_key_value_and_type()

        return Salesforce(
            username=self.username,
            consumer_key=self.access_config.consumer_key,
            privatekey_file=pkey_value if pkey_type is Path else None,
            privatekey=pkey_value if pkey_type is str else None,
            version=SALESFORCE_API_VERSION,
        )


@dataclass
class SalesforceIndexerConfig(IndexerConfig):
    categories: list[str]


@dataclass
class SalesforceIndexer(Indexer):
    connection_config: SalesforceConnectionConfig
    index_config: SalesforceIndexerConfig

    def __post_init__(self):
        for record_type in self.index_config.categories:
            if record_type not in ACCEPTED_CATEGORIES:
                raise ValueError(f"{record_type} not currently an accepted Salesforce category")

    def get_file_extension(self, record_type) -> str:
        if record_type == "EmailMessage":
            extension = ".eml"
        elif record_type in ["Account", "Lead", "Case", "Campaign"]:
            extension = ".xml"
        else:
            raise MissingCategoryError(
                f"There are no categories with the name: {record_type}",
            )
        return extension

    @requires_dependencies(["simple_salesforce"], extras="salesforce")
    def list_files(self) -> list[FileData]:
        """Get Salesforce Ids for the records.
        Send them to next phase where each doc gets downloaded into the
        appropriate format for partitioning.
        """
        from simple_salesforce.exceptions import SalesforceMalformedRequest

        client = self.connection_config.get_client()

        files_list = []
        for record_type in self.index_config.categories:
            try:
                # Get ids from Salesforce
                records = client.query_all_iter(
                    f"select Id, SystemModstamp, CreatedDate, LastModifiedDate from {record_type}",
                )
                for record in records:
                    record_with_extension = record["Id"] + self.get_file_extension(
                        record["attributes"]["type"]
                    )
                    files_list.append(
                        FileData(
                            connector_type=CONNECTOR_TYPE,
                            identifier=record["Id"],
                            source_identifiers=SourceIdentifiers(
                                filename=record_with_extension,
                                fullpath=f"{record['attributes']['type']}/{record_with_extension}",
                            ),
                            metadata=DataSourceMetadata(
                                url=record["attributes"]["url"],
                                version=str(parser.parse(record["SystemModstamp"]).timestamp()),
                                date_created=str(parser.parse(record["CreatedDate"]).timestamp()),
                                date_modified=str(
                                    parser.parse(record["LastModifiedDate"]).timestamp()
                                ),
                                record_locator={"id": record["Id"]},
                            ),
                            additional_metadata={"record_type": record["attributes"]["type"]},
                        )
                    )
            except SalesforceMalformedRequest as e:
                raise SalesforceMalformedRequest(f"Problem with Salesforce query: {e}")

        return files_list

    def run(self, **kwargs: Any) -> Generator[FileData, None, None]:
        for f in self.list_files():
            yield f


@dataclass
class SalesforceDownloaderConfig(DownloaderConfig):
    pass


@dataclass
class SalesforceDownloader(Downloader):
    connection_config: SalesforceConnectionConfig
    download_config: SalesforceDownloaderConfig = field(
        default_factory=lambda: SalesforceDownloaderConfig()
    )
    connector_type: str = CONNECTOR_TYPE

    def get_download_path(self, file_data: FileData) -> Path:
        rel_path = file_data.source_identifiers.relative_path
        rel_path = rel_path[1:] if rel_path.startswith("/") else rel_path
        return self.download_dir / Path(rel_path)

    def _xml_for_record(self, record: OrderedDict) -> str:
        """Creates partitionable xml file from a record"""
        import xml.etree.ElementTree as ET

        def create_xml_doc(data, parent, prefix=""):
            for key, value in data.items():
                if isinstance(value, OrderedDict):
                    create_xml_doc(value, parent, prefix=f"{prefix}{key}.")
                else:
                    item = ET.Element("item")
                    item.text = f"{prefix}{key}: {value}"
                    parent.append(item)

        root = ET.Element("root")
        create_xml_doc(record, root)

        xml_string = ET.tostring(root, encoding="utf-8", xml_declaration=True).decode()
        return xml_string

    def _eml_for_record(self, email_json: dict[str, Any]) -> str:
        """Recreates standard expected .eml format using template."""
        eml = EMAIL_TEMPLATE.substitute(
            date=formatdate(parser.parse(email_json.get("MessageDate")).timestamp()),
            message_identifier=email_json.get("MessageIdentifier"),
            subject=email_json.get("Subject"),
            from_email=email_json.get("FromAddress"),
            to_email=email_json.get("ToAddress"),
            textbody=email_json.get("TextBody"),
            htmlbody=email_json.get("HtmlBody"),
        )
        return dedent(eml)

    @SourceConnectionNetworkError.wrap
    def _get_response(self, file_data: FileData) -> OrderedDict:
        client = self.connection_config.get_client()
        return client.query(
            f"select FIELDS(STANDARD) from {file_data.additional_metadata['record_type']} where Id='{file_data.identifier}'",  # noqa: E501
        )

    def get_record(self, file_data: FileData) -> OrderedDict:
        # Get record from Salesforce based on id
        response = self._get_response(file_data)
        logger.debug(f"response was returned for salesforce record id: {file_data.identifier}")
        records = response["records"]
        if not records:
            raise ValueError(
                f"No record found with record id {file_data.identifier}: {json.dumps(response)}"
            )
        record_json = records[0]
        return record_json

    def run(self, file_data: FileData, **kwargs: Any) -> DownloadResponse:
        record = self.get_record(file_data)

        try:
            if file_data.additional_metadata["record_type"] == "EmailMessage":
                document = self._eml_for_record(record)
            else:
                document = self._xml_for_record(record)
            download_path = self.get_download_path(file_data=file_data)
            download_path.parent.mkdir(parents=True, exist_ok=True)

            with open(download_path, "w") as page_file:
                page_file.write(document)

        except Exception as e:
            logger.error(f"failed to download file {file_data.identifier}: {e}", exc_info=True)
            raise SourceConnectionNetworkError(f"failed to download file {file_data.identifier}")

        return self.generate_download_response(file_data=file_data, download_path=download_path)


salesforce_source_entry = SourceRegistryEntry(
    connection_config=SalesforceConnectionConfig,
    indexer_config=SalesforceIndexerConfig,
    indexer=SalesforceIndexer,
    downloader_config=SalesforceDownloaderConfig,
    downloader=SalesforceDownloader,
)



from dataclasses import dataclass

import click

from unstructured_ingest.v2.cli.base import DestCmd, SrcCmd
from unstructured_ingest.v2.cli.cmds.fsspec.fsspec import (
    FsspecCliDownloadConfig,
    FsspecCliIndexerConfig,
    FsspecCliUploaderConfig,
)
from unstructured_ingest.v2.cli.interfaces import CliConfig
from unstructured_ingest.v2.cli.utils import FileOrJson
from unstructured_ingest.v2.processes.connectors.fsspec.gcs import (
    CONNECTOR_TYPE,
)


@dataclass
class GcsCliDownloadConfig(FsspecCliDownloadConfig):
    pass


@dataclass
class GcsCliIndexerConfig(FsspecCliIndexerConfig):
    pass


@dataclass
class GcsCliConnectionConfig(CliConfig):
    @staticmethod
    def get_cli_options() -> list[click.Option]:
        help_string = """
          Options:
          - ``None``, GCSFS will attempt to guess your credentials in the
        following order: gcloud CLI default, gcsfs cached token, google compute
        metadata service, anonymous.
          - ``'google_default'``, your default gcloud credentials will be used,
            which are typically established by doing ``gcloud login`` in a terminal.
          - ``'cache'``, credentials from previously successful gcsfs
            authentication will be used (use this after "browser" auth succeeded)
          - ``'anon'``, no authentication is performed, and you can only
            access data which is accessible to allUsers (in this case, the project and
            access level parameters are meaningless)
          - ``'browser'``, you get an access code with which you can
            authenticate via a specially provided URL
          - if ``'cloud'``, we assume we are running within google compute
            or google container engine, and query the internal metadata directly for
            a token.
          - you may supply a token generated by the
            [gcloud](https://cloud.google.com/sdk/docs/)
            utility; this is either a python dictionary or the name of a file
            containing the JSON returned by logging in with the gcloud CLI tool.
          """
        options = [
            click.Option(
                ["--service-account-key"],
                default=None,
                type=FileOrJson(allow_raw_str=True),
                help=help_string,
            ),
        ]
        return options


@dataclass
class GcsUploaderConfig(FsspecCliUploaderConfig):
    pass


gcs_src_cmd = SrcCmd(
    cmd_name=CONNECTOR_TYPE,
    indexer_config=GcsCliIndexerConfig,
    connection_config=GcsCliConnectionConfig,
    downloader_config=GcsCliDownloadConfig,
)

gcs_dest_cmd = DestCmd(
    cmd_name=CONNECTOR_TYPE,
    connection_config=GcsCliConnectionConfig,
    uploader_config=GcsUploaderConfig,
)



import json
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional

from unstructured.__version__ import __version__ as unstructured_version

from unstructured_ingest.enhanced_dataclass import enhanced_field
from unstructured_ingest.utils.data_prep import batch_generator
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.interfaces import (
    AccessConfig,
    ConnectionConfig,
    FileData,
    UploadContent,
    Uploader,
    UploaderConfig,
    UploadStager,
    UploadStagerConfig,
)
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.processes.connector_registry import (
    DestinationRegistryEntry,
)

if TYPE_CHECKING:
    from pymongo import MongoClient

CONNECTOR_TYPE = "mongodb"
SERVER_API_VERSION = "1"


@dataclass
class MongoDBAccessConfig(AccessConfig):
    uri: Optional[str] = None


@dataclass
class MongoDBConnectionConfig(ConnectionConfig):
    access_config: MongoDBAccessConfig = enhanced_field(
        sensitive=True, default_factory=MongoDBAccessConfig
    )
    host: Optional[str] = None
    database: Optional[str] = None
    collection: Optional[str] = None
    port: int = 27017
    batch_size: int = 100
    connector_type: str = CONNECTOR_TYPE


@dataclass
class MongoDBUploadStagerConfig(UploadStagerConfig):
    pass


@dataclass
class MongoDBUploadStager(UploadStager):
    upload_stager_config: MongoDBUploadStagerConfig = field(
        default_factory=lambda: MongoDBUploadStagerConfig()
    )

    def run(
        self,
        elements_filepath: Path,
        file_data: FileData,
        output_dir: Path,
        output_filename: str,
        **kwargs: Any,
    ) -> Path:
        with open(elements_filepath) as elements_file:
            elements_contents = json.load(elements_file)

        output_path = Path(output_dir) / Path(f"{output_filename}.json")
        with open(output_path, "w") as output_file:
            json.dump(elements_contents, output_file)
        return output_path


@dataclass
class MongoDBUploaderConfig(UploaderConfig):
    batch_size: int = 100


@dataclass
class MongoDBUploader(Uploader):
    upload_config: MongoDBUploaderConfig
    connection_config: MongoDBConnectionConfig
    client: Optional["MongoClient"] = field(init=False)
    connector_type: str = CONNECTOR_TYPE

    def __post_init__(self):
        self.client = self.create_client()

    @requires_dependencies(["pymongo"], extras="mongodb")
    def create_client(self) -> "MongoClient":
        from pymongo import MongoClient
        from pymongo.driver_info import DriverInfo
        from pymongo.server_api import ServerApi

        if self.connection_config.access_config.uri:
            return MongoClient(
                self.connection_config.access_config.uri,
                server_api=ServerApi(version=SERVER_API_VERSION),
                driver=DriverInfo(name="unstructured", version=unstructured_version),
            )
        else:
            return MongoClient(
                host=self.connection_config.host,
                port=self.connection_config.port,
                server_api=ServerApi(version=SERVER_API_VERSION),
            )

    def run(self, contents: list[UploadContent], **kwargs: Any) -> None:
        elements_dict = []
        for content in contents:
            with open(content.path) as elements_file:
                elements = json.load(elements_file)
                elements_dict.extend(elements)

        logger.info(
            f"writing {len(elements_dict)} objects to destination "
            f"db, {self.connection_config.database}, "
            f"collection {self.connection_config.collection} "
            f"at {self.connection_config.host}",
        )
        db = self.client[self.connection_config.database]
        collection = db[self.connection_config.collection]
        for chunk in batch_generator(elements_dict, self.upload_config.batch_size):
            collection.insert_many(chunk)


mongodb_destination_entry = DestinationRegistryEntry(
    connection_config=MongoDBConnectionConfig,
    uploader=MongoDBUploader,
    uploader_config=MongoDBUploaderConfig,
    upload_stager=MongoDBUploadStager,
    upload_stager_config=MongoDBUploadStagerConfig,
)



import hashlib
import json
import typing as t
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from unstructured_ingest.connector.registry import create_ingest_doc_from_dict
from unstructured_ingest.error import PartitionError
from unstructured_ingest.logger import logger
from unstructured_ingest.pipeline.interfaces import PartitionNode
from unstructured_ingest.pipeline.utils import get_ingest_doc_hash


@dataclass
class Partitioner(PartitionNode):
    @PartitionError.wrap
    def run(self, ingest_doc_dict) -> Optional[str]:
        try:
            doc = create_ingest_doc_from_dict(ingest_doc_dict)
            doc_filename_hash = get_ingest_doc_hash(ingest_doc_dict)
            hashed_filename = hashlib.sha256(
                f"{self.create_hash()}{doc_filename_hash}".encode(),
            ).hexdigest()[:32]
            self.pipeline_context.ingest_docs_map[hashed_filename] = ingest_doc_dict
            doc_filename = f"{hashed_filename}.json"
            json_path = (Path(self.get_path()) / doc_filename).resolve()
            if (
                not self.pipeline_context.reprocess
                and json_path.is_file()
                and json_path.stat().st_size
            ):
                logger.info(f"File exists: {json_path}, skipping partition")
                return str(json_path)
            partition_kwargs: t.Dict[str, t.Any] = {
                "strategy": self.partition_config.strategy,
                "encoding": self.partition_config.encoding,
                "pdf_infer_table_structure": self.partition_config.pdf_infer_table_structure,
                "languages": self.partition_config.ocr_languages,
                "hi_res_model_name": self.partition_config.hi_res_model_name,
            }
            if self.partition_config.skip_infer_table_types:
                partition_kwargs["skip_infer_table_types"] = (
                    self.partition_config.skip_infer_table_types
                )
            if self.partition_config.additional_partition_args:
                partition_kwargs.update(self.partition_config.additional_partition_args)
            elements = doc.process_file(
                partition_config=self.partition_config,
                **partition_kwargs,
            )
            with open(json_path, "w", encoding="utf8") as output_f:
                logger.info(f"writing partitioned content to {json_path}")
                json.dump(elements, output_f, ensure_ascii=False, indent=2, sort_keys=True)
            return str(json_path)
        except Exception as e:
            if self.pipeline_context.raise_on_error:
                raise
            logger.error(f"failed to partition doc: {ingest_doc_dict}, {e}", exc_info=True)
            return None



from __future__ import annotations

import hashlib
import json
import os.path
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from unstructured.chunking import dispatch
from unstructured.documents.elements import Element, assign_and_map_hash_ids
from unstructured.partition.api import partition_via_api
from unstructured.staging.base import elements_from_json, elements_to_dicts

from unstructured_ingest.interfaces import ChunkingConfig, PartitionConfig
from unstructured_ingest.logger import logger
from unstructured_ingest.pipeline.interfaces import ReformatNode


@dataclass
class Chunker(ReformatNode):
    """Implementation for the chunking node in the ingest pipeline.

    Parameters
    ----------
    pipeline_context: PipelineContext (inherited from parent class)
    chunking_config: ChunkingConfig
    partition_config: PartitionConfig
    """

    chunking_config: ChunkingConfig
    partition_config: PartitionConfig

    def initialize(self):
        logger.info(
            f"Running chunking node. Chunking config: {self.chunking_config.to_json()}]",
        )
        super().initialize()

    def create_hash(self) -> str:
        hash_dict = self.chunking_config.to_dict()
        return hashlib.sha256(json.dumps(hash_dict, sort_keys=True).encode()).hexdigest()[:32]

    def run(self, elements_json: str) -> Optional[str]:
        try:
            elements_json_filename = os.path.basename(elements_json)
            filename_ext = os.path.basename(elements_json_filename)
            filename = os.path.splitext(filename_ext)[0]
            hashed_filename = hashlib.sha256(
                f"{self.create_hash()}{filename}".encode(),
            ).hexdigest()[:32]
            json_filename = f"{hashed_filename}.json"
            json_path = (Path(self.get_path()) / json_filename).resolve()
            self.pipeline_context.ingest_docs_map[hashed_filename] = (
                self.pipeline_context.ingest_docs_map[filename]
            )
            if (
                not self.pipeline_context.reprocess
                and json_path.is_file()
                and json_path.stat().st_size
            ):
                logger.debug(f"File exists: {json_path}, skipping chunking")
                return str(json_path)

            chunked_elements = self.chunk(elements_json)

            # -- return if chunking_strategy is None --
            if chunked_elements is None:
                logger.info(f"chunking_strategy is None, skipping chunking for {filename_ext}")
                return

            assign_and_map_hash_ids(chunked_elements)

            element_dicts = elements_to_dicts(chunked_elements)
            with open(json_path, "w", encoding="utf8") as output_f:
                logger.info(f"writing chunking content to {json_path}")
                json.dump(element_dicts, output_f, ensure_ascii=False, indent=2)
                return str(json_path)

        except Exception as e:
            if self.pipeline_context.raise_on_error:
                raise
            logger.error(f"failed to run chunking on file {elements_json}, {e}", exc_info=True)
            return None

    def get_path(self) -> Path:
        return (Path(self.pipeline_context.work_dir) / "chunked").resolve()

    def chunk(self, elements_json_file: str) -> Optional[list[Element]]:
        """Called by Chunker.run() to properly execute the defined chunking_strategy."""
        # -- No chunking_strategy means no chunking --
        if self.chunking_config.chunking_strategy is None:
            return
        # -- Chunk locally for open-source chunking strategies, even when partitioning remotely --
        if self.chunking_config.chunking_strategy in ("basic", "by_title"):
            return dispatch.chunk(
                elements=elements_from_json(filename=elements_json_file),
                chunking_strategy=self.chunking_config.chunking_strategy,
                combine_text_under_n_chars=self.chunking_config.combine_text_under_n_chars,
                include_orig_elements=self.chunking_config.include_orig_elements,
                max_characters=self.chunking_config.max_characters,
                multipage_sections=self.chunking_config.multipage_sections,
                new_after_n_chars=self.chunking_config.new_after_n_chars,
                overlap=self.chunking_config.overlap,
                overlap_all=self.chunking_config.overlap_all,
            )
        # -- Chunk remotely --
        if self.partition_config.partition_by_api:
            return partition_via_api(
                filename=elements_json_file,
                # -- (jennings) If api_key or api_url are None, partition_via_api will raise an
                # -- error, which will be caught and logged by Chunker.run()
                api_key=self.partition_config.api_key,  # type: ignore
                api_url=self.partition_config.partition_endpoint,  # type: ignore
                chunking_strategy=self.chunking_config.chunking_strategy,
                combine_under_n_chars=self.chunking_config.combine_text_under_n_chars,
                include_orig_elements=self.chunking_config.include_orig_elements,
                max_characters=self.chunking_config.max_characters,
                multipage_sections=self.chunking_config.multipage_sections,
                new_after_n_chars=self.chunking_config.new_after_n_chars,
                overlap=self.chunking_config.overlap,
                overlap_all=self.chunking_config.overlap_all,
            )
        # -- Warn that the defined chunking_strategy is not locally available --
        logger.warning(
            f"There is no locally available chunking_strategy:"
            f" {self.chunking_config.chunking_strategy}."
            f" If trying to partition remotely, check that `partition_by_api`, `api_url`,"
            f" and `api_key` are correctly defined."
        )



from datetime import datetime
from typing import Union

from dateutil import parser


def parse_datetime(date_value: Union[int, str, float, datetime]) -> datetime:
    if isinstance(date_value, datetime):
        return date_value
    elif isinstance(date_value, float):
        return datetime.fromtimestamp(date_value)
    elif isinstance(date_value, int):
        return datetime.fromtimestamp(date_value / 1000)

    try:
        timestamp = float(date_value)
        return datetime.fromtimestamp(timestamp)
    except ValueError:
        return parser.parse(date_value)



import copy
import hashlib
import typing as t
import uuid
from dataclasses import dataclass, field
from pathlib import Path

from dataclasses_json.core import Json

from unstructured_ingest.enhanced_dataclass import enhanced_field
from unstructured_ingest.enhanced_dataclass.core import _asdict
from unstructured_ingest.error import DestinationConnectionError, SourceConnectionError
from unstructured_ingest.interfaces import (
    AccessConfig,
    BaseConnectorConfig,
    BaseDestinationConnector,
    BaseIngestDocBatch,
    BaseSingleIngestDoc,
    BaseSourceConnector,
    IngestDocCleanupMixin,
    SourceConnectorCleanupMixin,
    SourceMetadata,
    WriteConfig,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.data_prep import flatten_dict, generator_batching_wbytes
from unstructured_ingest.utils.dep_check import requires_dependencies

if t.TYPE_CHECKING:
    from elasticsearch import Elasticsearch


@dataclass
class ElasticsearchAccessConfig(AccessConfig):
    hosts: t.Optional[t.List[str]] = None
    username: t.Optional[str] = None
    password: t.Optional[str] = enhanced_field(default=None, sensitive=True)
    cloud_id: t.Optional[str] = None
    api_key: t.Optional[str] = enhanced_field(
        default=None, sensitive=True, overload_name="es_api_key"
    )
    api_key_id: t.Optional[str] = None
    bearer_auth: t.Optional[str] = enhanced_field(default=None, sensitive=True)
    ca_certs: t.Optional[str] = None
    ssl_assert_fingerprint: t.Optional[str] = enhanced_field(default=None, sensitive=True)

    def to_dict(self, **kwargs) -> t.Dict[str, Json]:
        d = super().to_dict(**kwargs)
        # Update auth related fields to conform to what the SDK expects based on the
        # supported methods:
        # https://www.elastic.co/guide/en/elasticsearch/client/python-api/current/connecting.html
        if not self.ca_certs:
            # ES library already sets a default for this, don't want to
            # introduce data by setting it to None
            d.pop("ca_certs")
        if self.password and (self.cloud_id or self.ca_certs or self.ssl_assert_fingerprint):
            d.pop("password")
            d["basic_auth"] = ("elastic", self.password)
        elif not self.cloud_id and self.username and self.password:
            d.pop("username", None)
            d.pop("password", None)
            d["basic_auth"] = (self.username, self.password)
        elif self.api_key and self.api_key_id:
            d.pop("api_key_id", None)
            d.pop("api_key", None)
            d["api_key"] = (self.api_key_id, self.api_key)
        # This doesn't exist on the client init, remove:
        d.pop("api_key_id", None)
        return d


@dataclass
class SimpleElasticsearchConfig(BaseConnectorConfig):
    """Connector config where:
    url is the url to access the elasticsearch server,
    index_name is the name of the index to reach to,
    """

    index_name: str
    batch_size: int = 100
    fields: t.List[str] = field(default_factory=list)
    access_config: ElasticsearchAccessConfig = None


@dataclass
class ElasticsearchDocumentMeta:
    """Metadata specifying:
    name of the elasticsearch index that is being reached to,
    and the id of document that is being reached to,
    """

    index_name: str
    document_id: str


@dataclass
class ElasticsearchIngestDoc(IngestDocCleanupMixin, BaseSingleIngestDoc):
    """Class encapsulating fetching a doc and writing processed results (but not
    doing the processing!).

    Current implementation creates a python Elasticsearch client to fetch each doc,
    rather than creating a client for each thread.
    """

    connector_config: SimpleElasticsearchConfig
    document_meta: ElasticsearchDocumentMeta
    document: dict = field(default_factory=dict)
    registry_name: str = "elasticsearch"

    # TODO: remove one of filename or _tmp_download_file, using a wrapper
    @property
    def filename(self):
        f = self.document_meta.document_id
        if self.connector_config.fields:
            f = "{}-{}".format(
                f,
                hashlib.sha256(",".join(self.connector_config.fields).encode()).hexdigest()[:8],
            )
        return (
            Path(self.read_config.download_dir) / self.document_meta.index_name / f"{f}.txt"
        ).resolve()

    @property
    def _output_filename(self):
        """Create filename document id combined with a hash of the query to uniquely identify
        the output file."""
        # Generate SHA256 hash and take the first 8 characters
        filename = self.document_meta.document_id
        if self.connector_config.fields:
            filename = "{}-{}".format(
                filename,
                hashlib.sha256(",".join(self.connector_config.fields).encode()).hexdigest()[:8],
            )
        output_file = f"{filename}.json"
        return (
            Path(self.processor_config.output_dir) / self.connector_config.index_name / output_file
        )

    def update_source_metadata(self, **kwargs):
        if self.document is None:
            self.source_metadata = SourceMetadata(
                exists=False,
            )
            return
        self.source_metadata = SourceMetadata(
            version=self.document["_version"],
            exists=True,
        )

    @SourceConnectionError.wrap
    @requires_dependencies(["elasticsearch"], extras="elasticsearch")
    @BaseSingleIngestDoc.skip_if_file_exists
    def get_file(self):
        pass

    @property
    def date_created(self) -> t.Optional[str]:
        return None

    @property
    def date_modified(self) -> t.Optional[str]:
        return None

    @property
    def source_url(self) -> t.Optional[str]:
        return None

    @property
    def record_locator(self) -> t.Optional[t.Dict[str, t.Any]]:
        return {
            "hosts": self.connector_config.access_config.hosts,
            "index_name": self.connector_config.index_name,
            "document_id": self.document_meta.document_id,
        }


@dataclass
class ElasticsearchIngestDocBatch(BaseIngestDocBatch):
    connector_config: SimpleElasticsearchConfig
    ingest_docs: t.List[ElasticsearchIngestDoc] = field(default_factory=list)
    list_of_ids: t.List[str] = field(default_factory=list)
    registry_name: str = "elasticsearch_batch"

    def __post_init__(self):
        # Until python3.8 is deprecated, this is a limitation of dataclass inheritance
        # to make it a required field
        if len(self.list_of_ids) == 0:
            raise ValueError("list_of_ids is required")

    @property
    def unique_id(self) -> str:
        return ",".join(sorted(self.list_of_ids))

    @requires_dependencies(["elasticsearch"], extras="elasticsearch")
    def _get_docs(self):
        from elasticsearch import Elasticsearch
        from elasticsearch.helpers import scan

        es = Elasticsearch(**self.connector_config.access_config.to_dict(apply_name_overload=False))
        scan_query = {
            "_source": self.connector_config.fields,
            "version": True,
            "query": {"ids": {"values": self.list_of_ids}},
        }

        result = scan(
            es,
            query=scan_query,
            scroll="1m",
            index=self.connector_config.index_name,
        )
        return list(result)

    @SourceConnectionError.wrap
    @requires_dependencies(["elasticsearch"], extras="elasticsearch")
    def get_files(self):
        documents = self._get_docs()
        for doc in documents:
            ingest_doc = ElasticsearchIngestDoc(
                processor_config=self.processor_config,
                read_config=self.read_config,
                connector_config=self.connector_config,
                document=doc,
                document_meta=ElasticsearchDocumentMeta(
                    self.connector_config.index_name, doc["_id"]
                ),
            )
            ingest_doc.update_source_metadata()
            doc_body = doc["_source"]
            filename = ingest_doc.filename
            flattened_dict = flatten_dict(dictionary=doc_body)
            str_values = [str(value) for value in flattened_dict.values()]
            concatenated_values = "\n".join(str_values)

            filename.parent.mkdir(parents=True, exist_ok=True)
            with open(filename, "w", encoding="utf8") as f:
                f.write(concatenated_values)
            self.ingest_docs.append(ingest_doc)


@dataclass
class ElasticsearchSourceConnector(SourceConnectorCleanupMixin, BaseSourceConnector):
    """Fetches particular fields from all documents in a given elasticsearch cluster and index"""

    connector_config: SimpleElasticsearchConfig
    _es: t.Optional["Elasticsearch"] = field(init=False, default=None)

    @property
    def es(self):
        from elasticsearch import Elasticsearch

        if self._es is None:
            self._es = Elasticsearch(
                **self.connector_config.access_config.to_dict(apply_name_overload=False)
            )
        return self._es

    def check_connection(self):
        try:
            self.es.perform_request("HEAD", "/", headers={"accept": "application/json"})
        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise SourceConnectionError(f"failed to validate connection: {e}")

    def __post_init__(self):
        self.scan_query: dict = {"stored_fields": [], "query": {"match_all": {}}}

    def initialize(self):
        pass

    @requires_dependencies(["elasticsearch"], extras="elasticsearch")
    def _get_doc_ids(self):
        """Fetches all document ids in an index"""
        from elasticsearch.helpers import scan

        hits = scan(
            self.es,
            query=self.scan_query,
            scroll="1m",
            index=self.connector_config.index_name,
        )

        return [hit["_id"] for hit in hits]

    def get_ingest_docs(self):
        """Fetches all documents in an index, using ids that are fetched with _get_doc_ids"""
        ids = self._get_doc_ids()
        id_batches = [
            ids[
                i
                * self.connector_config.batch_size : (i + 1)  # noqa
                * self.connector_config.batch_size
            ]
            for i in range(
                (len(ids) + self.connector_config.batch_size - 1)
                // self.connector_config.batch_size
            )
        ]
        return [
            ElasticsearchIngestDocBatch(
                connector_config=self.connector_config,
                processor_config=self.processor_config,
                read_config=self.read_config,
                list_of_ids=batched_ids,
            )
            for batched_ids in id_batches
        ]


@dataclass
class ElasticsearchWriteConfig(WriteConfig):
    batch_size_bytes: int = 15_000_000
    num_processes: int = 1


@dataclass
class ElasticsearchDestinationConnector(BaseDestinationConnector):
    write_config: ElasticsearchWriteConfig
    connector_config: SimpleElasticsearchConfig
    _client: t.Optional["Elasticsearch"] = field(init=False, default=None)

    def to_dict(self, **kwargs):
        """
        The _client variable in this dataclass breaks deepcopy due to:
        TypeError: cannot pickle '_thread.lock' object
        When serializing, remove it, meaning client data will need to be reinitialized
        when deserialized
        """
        self_cp = copy.copy(self)
        if hasattr(self_cp, "_client"):
            setattr(self_cp, "_client", None)
        return _asdict(self_cp, **kwargs)

    @DestinationConnectionError.wrap
    @requires_dependencies(["elasticsearch"], extras="elasticsearch")
    def generate_client(self) -> "Elasticsearch":
        from elasticsearch import Elasticsearch

        return Elasticsearch(
            **self.connector_config.access_config.to_dict(apply_name_overload=False)
        )

    @property
    def client(self):
        if self._client is None:
            self._client = self.generate_client()
        return self._client

    def initialize(self):
        _ = self.client

    @DestinationConnectionError.wrap
    def check_connection(self):
        try:
            assert self.client.ping()
        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise DestinationConnectionError(f"failed to validate connection: {e}")

    @requires_dependencies(["elasticsearch"], extras="elasticsearch")
    def write_dict(self, *args, elements_dict: t.List[t.Dict[str, t.Any]], **kwargs) -> None:
        logger.info(
            f"writing document batches to destination"
            f" index named {self.connector_config.index_name}"
            f" at {self.connector_config.access_config.hosts}"
            f" with batch size (in bytes) {self.write_config.batch_size_bytes}"
            f" with {self.write_config.num_processes} (number of) processes"
        )
        from elasticsearch.helpers import parallel_bulk

        for batch in generator_batching_wbytes(
            elements_dict, batch_size_limit_bytes=self.write_config.batch_size_bytes
        ):
            for success, info in parallel_bulk(
                self.client, batch, thread_count=self.write_config.num_processes
            ):
                if not success:
                    logger.error(
                        "upload failed for a batch in elasticsearch destination connector:", info
                    )

    def normalize_dict(self, element_dict: dict) -> dict:
        return {
            "_index": self.connector_config.index_name,
            "_id": str(uuid.uuid4()),
            "_source": {
                "element_id": element_dict.pop("element_id", None),
                "embeddings": element_dict.pop("embeddings", None),
                "text": element_dict.pop("text", None),
                "type": element_dict.pop("type", None),
                "metadata": flatten_dict(
                    element_dict.pop("metadata", None),
                    separator="-",
                ),
            },
        }



from dataclasses import dataclass

import click

from unstructured_ingest.v2.cli.base import DestCmd, SrcCmd
from unstructured_ingest.v2.cli.cmds.fsspec.fsspec import (
    FsspecCliDownloadConfig,
    FsspecCliIndexerConfig,
    FsspecCliUploaderConfig,
)
from unstructured_ingest.v2.cli.interfaces import CliConfig
from unstructured_ingest.v2.processes.connectors.fsspec.dropbox import (
    CONNECTOR_TYPE,
)


@dataclass
class DropboxCliDownloadConfig(FsspecCliDownloadConfig):
    pass


@dataclass
class DropboxCliIndexerConfig(FsspecCliIndexerConfig):
    pass


@dataclass
class DropboxCliConnectionConfig(CliConfig):
    @staticmethod
    def get_cli_options() -> list[click.Option]:
        options = [
            click.Option(
                ["--token"],
                required=True,
                type=str,
                help="Dropbox access token.",
            ),
        ]
        return options


@dataclass
class DropboxUploaderConfig(FsspecCliUploaderConfig):
    pass


dropbox_src_cmd = SrcCmd(
    cmd_name=CONNECTOR_TYPE,
    indexer_config=DropboxCliIndexerConfig,
    connection_config=DropboxCliConnectionConfig,
    downloader_config=DropboxCliDownloadConfig,
)

dropbox_dest_cmd = DestCmd(
    cmd_name=CONNECTOR_TYPE,
    connection_config=DropboxCliConnectionConfig,
    uploader_config=DropboxUploaderConfig,
)



from dataclasses import dataclass

import click

from unstructured_ingest.v2.cli.base import SrcCmd
from unstructured_ingest.v2.cli.interfaces import CliConfig
from unstructured_ingest.v2.processes.connectors.sharepoint import CONNECTOR_TYPE


@dataclass
class SharepointCliConnectionConfig(CliConfig):
    @staticmethod
    def get_cli_options() -> list[click.Option]:
        options = [
            click.Option(
                ["--client-id"],
                default=None,
                type=str,
                help="Sharepoint app client ID",
            ),
            click.Option(
                ["--client-cred"],
                default=None,
                type=str,
                help="Sharepoint app secret",
            ),
            click.Option(
                ["--site"],
                default=None,
                type=str,
                help="Sharepoint site url. Process either base url e.g \
                    https://[tenant].sharepoint.com  or relative sites \
                    https://[tenant].sharepoint.com/sites/<site_name>. \
                    To process all sites within the tenant pass a site url as \
                    https://[tenant]-admin.sharepoint.com.\
                    This requires the app to be registered at a tenant level",
            ),
            click.Option(
                ["--permissions-application-id"],
                type=str,
                help="Microsoft Graph API application id",
            ),
            click.Option(
                ["--permissions-client-cred"],
                type=str,
                help="Microsoft Graph API application credentials",
            ),
            click.Option(
                ["--permissions-tenant"],
                type=str,
                help="e.g https://contoso.onmicrosoft.com to get permissions data within tenant.",
            ),
        ]
        return options


@dataclass
class SharepointCliIndexerConfig(CliConfig):
    @staticmethod
    def get_cli_options() -> list[click.Option]:
        options = [
            click.Option(
                ["--path"],
                default=None,
                type=str,
                help="Path from which to start parsing files. If the connector is to \
                process all sites within the tenant this filter will be applied to \
                all sites document libraries.",
            ),
            click.Option(
                ["--recursive"],
                is_flag=True,
                default=False,
                help="Recursively download files in their respective folders "
                "otherwise stop at the files in provided folder level.",
            ),
            click.Option(
                ["--omit-files"],
                is_flag=True,
                default=False,
                help="Don't process files.",
            ),
            click.Option(
                ["--omit-pages"],
                is_flag=True,
                default=False,
                help="Don't process site pages.",
            ),
        ]
        return options


@dataclass
class SharepointCliDownloadConfig(CliConfig):
    @staticmethod
    def get_cli_options() -> list[click.Option]:
        options = [
            click.Option(
                ["--download-dir"],
                help="Where files are downloaded to, defaults to a location at"
                "`$HOME/.cache/unstructured_ingest/<connector name>/<SHA256>`.",
            ),
        ]
        return options


sharepoint_drive_src_cmd = SrcCmd(
    cmd_name=CONNECTOR_TYPE,
    connection_config=SharepointCliConnectionConfig,
    indexer_config=SharepointCliIndexerConfig,
    downloader_config=SharepointCliDownloadConfig,
)



import os
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Optional

from unstructured_ingest.enhanced_dataclass import enhanced_field
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.interfaces import (
    AccessConfig,
    ConnectionConfig,
    UploadContent,
    Uploader,
    UploaderConfig,
)
from unstructured_ingest.v2.processes.connector_registry import DestinationRegistryEntry

if TYPE_CHECKING:
    from databricks.sdk import WorkspaceClient

CONNECTOR_TYPE = "databricks_volumes"


@dataclass
class DatabricksVolumesAccessConfig(AccessConfig):
    account_id: Optional[str] = None
    username: Optional[str] = None
    password: Optional[str] = None
    client_id: Optional[str] = None
    client_secret: Optional[str] = None
    token: Optional[str] = None
    profile: Optional[str] = None
    azure_workspace_resource_id: Optional[str] = None
    azure_client_secret: Optional[str] = None
    azure_client_id: Optional[str] = None
    azure_tenant_id: Optional[str] = None
    azure_environment: Optional[str] = None
    auth_type: Optional[str] = None
    cluster_id: Optional[str] = None
    google_credentials: Optional[str] = None
    google_service_account: Optional[str] = None


@dataclass
class DatabricksVolumesConnectionConfig(ConnectionConfig):
    access_config: DatabricksVolumesAccessConfig = enhanced_field(
        default_factory=DatabricksVolumesAccessConfig, sensitive=True
    )
    host: Optional[str] = None


@dataclass
class DatabricksVolumesUploaderConfig(UploaderConfig):
    volume: str
    catalog: str
    volume_path: Optional[str] = None
    overwrite: bool = False
    schema: str = "default"

    @property
    def path(self) -> str:
        path = f"/Volumes/{self.catalog}/{self.schema}/{self.volume}"
        if self.volume_path:
            path = f"{path}/{self.volume_path}"
        return path


@dataclass
class DatabricksVolumesUploader(Uploader):
    connector_type: str = CONNECTOR_TYPE
    upload_config: DatabricksVolumesUploaderConfig
    connection_config: DatabricksVolumesConnectionConfig
    client: Optional["WorkspaceClient"] = field(init=False, default=None)

    @requires_dependencies(dependencies=["databricks.sdk"], extras="databricks-volumes")
    def __post_init__(self) -> "WorkspaceClient":
        from databricks.sdk import WorkspaceClient

        self.client = WorkspaceClient(
            host=self.connection_config.host, **self.connection_config.access_config.to_dict()
        )

    def run(self, contents: list[UploadContent], **kwargs: Any) -> None:
        for content in contents:
            with open(content.path, "rb") as elements_file:
                output_path = os.path.join(self.upload_config.path, content.path.name)
                self.client.files.upload(
                    file_path=output_path,
                    contents=elements_file,
                    overwrite=self.upload_config.overwrite,
                )


databricks_volumes_destination_entry = DestinationRegistryEntry(
    connection_config=DatabricksVolumesConnectionConfig,
    uploader=DatabricksVolumesUploader,
    uploader_config=DatabricksVolumesUploaderConfig,
)



import contextlib
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from time import time
from typing import Any, Generator, Optional

from unstructured.documents.elements import DataSourceMetadata

from unstructured_ingest.enhanced_dataclass import enhanced_field
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.interfaces import DownloadResponse, FileData, UploadContent
from unstructured_ingest.v2.processes.connector_registry import (
    DestinationRegistryEntry,
    SourceRegistryEntry,
)
from unstructured_ingest.v2.processes.connectors.fsspec.fsspec import (
    FsspecAccessConfig,
    FsspecConnectionConfig,
    FsspecDownloader,
    FsspecDownloaderConfig,
    FsspecIndexer,
    FsspecIndexerConfig,
    FsspecUploader,
    FsspecUploaderConfig,
)

CONNECTOR_TYPE = "s3"


@dataclass
class S3IndexerConfig(FsspecIndexerConfig):
    pass


@dataclass
class S3AccessConfig(FsspecAccessConfig):
    key: Optional[str] = None
    secret: Optional[str] = None
    token: Optional[str] = None


@dataclass
class S3ConnectionConfig(FsspecConnectionConfig):
    supported_protocols: list[str] = field(default_factory=lambda: ["s3", "s3a"])
    access_config: S3AccessConfig = enhanced_field(
        sensitive=True, default_factory=lambda: S3AccessConfig()
    )
    endpoint_url: Optional[str] = None
    anonymous: bool = False
    connector_type: str = CONNECTOR_TYPE

    def get_access_config(self) -> dict[str, Any]:
        access_configs: dict[str, Any] = {"anon": self.anonymous}
        if self.endpoint_url:
            access_configs["endpoint_url"] = self.endpoint_url

        # Avoid injecting None by filtering out k,v pairs where the value is None
        access_configs.update({k: v for k, v in self.access_config.to_dict().items() if v})
        return access_configs


@dataclass
class S3Indexer(FsspecIndexer):
    connection_config: S3ConnectionConfig
    index_config: S3IndexerConfig
    connector_type: str = CONNECTOR_TYPE

    def get_metadata(self, path: str) -> DataSourceMetadata:
        date_created = None
        date_modified = None
        try:
            modified: Optional[datetime] = self.fs.modified(path)
            if modified:
                date_created = str(modified.timestamp())
                date_modified = str(modified.timestamp())
        except NotImplementedError:
            pass

        version = None
        info: dict[str, Any] = self.fs.info(path)
        if etag := info.get("ETag"):
            version = str(etag).rstrip('"').lstrip('"')
        metadata: dict[str, str] = {}
        with contextlib.suppress(AttributeError):
            metadata = self.fs.metadata(path)
        record_locator = {
            "protocol": self.index_config.protocol,
            "remote_file_path": self.index_config.remote_url,
        }
        if metadata:
            record_locator["metadata"] = metadata
        return DataSourceMetadata(
            date_created=date_created,
            date_modified=date_modified,
            date_processed=str(time()),
            version=version,
            url=f"{self.index_config.protocol}://{path}",
            record_locator=record_locator,
        )

    @requires_dependencies(["s3fs", "fsspec"], extras="s3")
    def run(self, **kwargs: Any) -> Generator[FileData, None, None]:
        return super().run(**kwargs)


@dataclass
class S3DownloaderConfig(FsspecDownloaderConfig):
    pass


@dataclass
class S3Downloader(FsspecDownloader):
    protocol: str = "s3"
    connection_config: S3ConnectionConfig
    connector_type: str = CONNECTOR_TYPE
    download_config: Optional[S3DownloaderConfig] = field(default_factory=S3DownloaderConfig)

    @requires_dependencies(["s3fs", "fsspec"], extras="s3")
    def run(self, file_data: FileData, **kwargs: Any) -> DownloadResponse:
        return super().run(file_data=file_data, **kwargs)

    @requires_dependencies(["s3fs", "fsspec"], extras="s3")
    async def run_async(self, file_data: FileData, **kwargs: Any) -> DownloadResponse:
        return await super().run_async(file_data=file_data, **kwargs)


@dataclass
class S3UploaderConfig(FsspecUploaderConfig):
    pass


@dataclass
class S3Uploader(FsspecUploader):
    connector_type: str = CONNECTOR_TYPE
    connection_config: S3ConnectionConfig
    upload_config: S3UploaderConfig = field(default=None)

    @requires_dependencies(["s3fs", "fsspec"], extras="s3")
    def __post_init__(self):
        super().__post_init__()

    @requires_dependencies(["s3fs", "fsspec"], extras="s3")
    def run(self, contents: list[UploadContent], **kwargs: Any) -> None:
        return super().run(contents=contents, **kwargs)

    @requires_dependencies(["s3fs", "fsspec"], extras="s3")
    async def run_async(self, path: Path, file_data: FileData, **kwargs: Any) -> None:
        return await super().run_async(path=path, file_data=file_data, **kwargs)


s3_source_entry = SourceRegistryEntry(
    indexer=S3Indexer,
    indexer_config=S3IndexerConfig,
    downloader=S3Downloader,
    downloader_config=S3DownloaderConfig,
    connection_config=S3ConnectionConfig,
)

s3_destination_entry = DestinationRegistryEntry(
    uploader=S3Uploader,
    uploader_config=S3UploaderConfig,
    connection_config=S3ConnectionConfig,
)



import json
from dataclasses import dataclass, field
from pathlib import Path
from time import time
from typing import TYPE_CHECKING, Any, Generator, Optional

from dateutil import parser
from unstructured.documents.elements import DataSourceMetadata

from unstructured_ingest.enhanced_dataclass import enhanced_field
from unstructured_ingest.error import SourceConnectionError, SourceConnectionNetworkError
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Downloader,
    DownloaderConfig,
    DownloadResponse,
    FileData,
    Indexer,
    IndexerConfig,
    SourceIdentifiers,
    download_responses,
)
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.processes.connector_registry import (
    SourceRegistryEntry,
)

if TYPE_CHECKING:
    from office365.graph_client import GraphClient
    from office365.onedrive.driveitems.driveItem import DriveItem

CONNECTOR_TYPE = "onedrive"
MAX_MB_SIZE = 512_000_000


@dataclass
class OnedriveAccessConfig(AccessConfig):
    client_cred: str


@dataclass
class OnedriveConnectionConfig(ConnectionConfig):
    client_id: str
    user_pname: str
    tenant: str = field(repr=False)
    authority_url: Optional[str] = field(repr=False, default="https://login.microsoftonline.com")
    access_config: OnedriveAccessConfig = enhanced_field(sensitive=True)

    @requires_dependencies(["msal"], extras="onedrive")
    def get_token(self):
        from msal import ConfidentialClientApplication

        try:
            app = ConfidentialClientApplication(
                authority=f"{self.authority_url}/{self.tenant}",
                client_id=self.client_id,
                client_credential=self.access_config.client_cred,
            )
            token = app.acquire_token_for_client(scopes=["https://graph.microsoft.com/.default"])
        except ValueError as exc:
            logger.error("Couldn't set up credentials for OneDrive")
            raise exc
        if "error" in token:
            raise SourceConnectionNetworkError(
                "failed to fetch token, {}: {}".format(token["error"], token["error_description"])
            )
        return token

    @requires_dependencies(["office365"], extras="onedrive")
    def get_client(self) -> "GraphClient":
        from office365.graph_client import GraphClient

        client = GraphClient(self.get_token)
        return client


@dataclass
class OnedriveIndexerConfig(IndexerConfig):
    path: Optional[str] = field(default="")
    recursive: bool = False


@dataclass
class OnedriveIndexer(Indexer):
    connection_config: OnedriveConnectionConfig
    index_config: OnedriveIndexerConfig

    def list_objects(self, folder, recursive) -> list["DriveItem"]:
        drive_items = folder.children.get().execute_query()
        files = [d for d in drive_items if d.is_file]
        if not recursive:
            return files
        folders = [d for d in drive_items if d.is_folder]
        for f in folders:
            files.extend(self.list_objects(f, recursive))
        return files

    def get_root(self, client: "GraphClient") -> "DriveItem":
        root = client.users[self.connection_config.user_pname].drive.get().execute_query().root
        if fpath := self.index_config.path:
            root = root.get_by_path(fpath).get().execute_query()
            if root is None or not root.is_folder:
                raise ValueError(f"Unable to find directory, given: {fpath}")
        return root

    def get_properties(self, drive_item: "DriveItem") -> dict:
        properties = drive_item.properties
        filtered_properties = {}
        for k, v in properties.items():
            try:
                json.dumps(v)
                filtered_properties[k] = v
            except TypeError:
                pass
        return filtered_properties

    def drive_item_to_file_data(self, drive_item: "DriveItem") -> FileData:
        file_path = drive_item.parent_reference.path.split(":")[-1]
        file_path = file_path[1:] if file_path and file_path[0] == "/" else file_path
        filename = drive_item.name
        server_path = file_path + "/" + filename
        rel_path = server_path.replace(self.index_config.path, "").lstrip("/")
        date_modified_dt = (
            parser.parse(drive_item.last_modified_datetime)
            if drive_item.last_modified_datetime
            else None
        )
        date_created_at = (
            parser.parse(drive_item.created_datetime) if drive_item.created_datetime else None
        )
        return FileData(
            identifier=drive_item.id,
            connector_type=CONNECTOR_TYPE,
            source_identifiers=SourceIdentifiers(
                fullpath=server_path, filename=drive_item.name, rel_path=rel_path
            ),
            metadata=DataSourceMetadata(
                url=drive_item.parent_reference.path + "/" + drive_item.name,
                version=drive_item.etag,
                date_modified=str(date_modified_dt.timestamp()) if date_modified_dt else None,
                date_created=str(date_created_at.timestamp()) if date_modified_dt else None,
                date_processed=str(time()),
                record_locator={
                    "user_pname": self.connection_config.user_pname,
                    "server_relative_path": server_path,
                },
            ),
            additional_metadata=self.get_properties(drive_item=drive_item),
        )

    def run(self, **kwargs: Any) -> Generator[FileData, None, None]:
        client = self.connection_config.get_client()
        root = self.get_root(client=client)
        drive_items = self.list_objects(folder=root, recursive=self.index_config.recursive)
        for drive_item in drive_items:
            file_data = self.drive_item_to_file_data(drive_item=drive_item)
            yield file_data


@dataclass
class OnedriveDownloaderConfig(DownloaderConfig):
    pass


@dataclass
class OnedriveDownloader(Downloader):
    connection_config: OnedriveConnectionConfig
    download_config: OnedriveDownloaderConfig

    @SourceConnectionNetworkError.wrap
    def _fetch_file(self, file_data: FileData):
        if file_data.source_identifiers is None or not file_data.source_identifiers.fullpath:
            raise ValueError(
                f"file data doesn't have enough information to get "
                f"file content: {file_data.to_dict()}"
            )

        server_relative_path = file_data.source_identifiers.fullpath
        client = self.connection_config.get_client()
        root = client.users[self.connection_config.user_pname].drive.get().execute_query().root
        file = root.get_by_path(server_relative_path).get().execute_query()
        if not file:
            raise FileNotFoundError(f"file not found: {server_relative_path}")
        return file

    def get_download_path(self, file_data: FileData) -> Optional[Path]:
        rel_path = file_data.source_identifiers.relative_path
        rel_path = rel_path[1:] if rel_path.startswith("/") else rel_path
        return self.download_dir / Path(rel_path)

    @SourceConnectionError.wrap
    def run(self, file_data: FileData, **kwargs: Any) -> download_responses:
        file = self._fetch_file(file_data=file_data)
        fsize = file.get_property("size", 0)
        download_path = self.get_download_path(file_data=file_data)
        download_path.parent.mkdir(parents=True, exist_ok=True)
        logger.info(f"Downloading {file_data.source_identifiers.fullpath} to {download_path}")
        if fsize > MAX_MB_SIZE:
            logger.info(f"Downloading file with size: {fsize} bytes in chunks")
            with download_path.open(mode="wb") as f:
                file.download_session(f, chunk_size=1024 * 1024 * 100).execute_query()
        else:
            with download_path.open(mode="wb") as f:
                file.download(f).execute_query()
        return DownloadResponse(file_data=file_data, path=download_path)


onedrive_source_entry = SourceRegistryEntry(
    connection_config=OnedriveConnectionConfig,
    indexer_config=OnedriveIndexerConfig,
    indexer=OnedriveIndexer,
    downloader_config=OnedriveDownloaderConfig,
    downloader=OnedriveDownloader,
)



from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Generator, Optional

from unstructured_ingest.enhanced_dataclass import enhanced_field
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.interfaces import DownloadResponse, FileData, UploadContent
from unstructured_ingest.v2.processes.connector_registry import (
    DestinationRegistryEntry,
    SourceRegistryEntry,
)
from unstructured_ingest.v2.processes.connectors.fsspec.fsspec import (
    FsspecAccessConfig,
    FsspecConnectionConfig,
    FsspecDownloader,
    FsspecDownloaderConfig,
    FsspecIndexer,
    FsspecIndexerConfig,
    FsspecUploader,
    FsspecUploaderConfig,
)
from unstructured_ingest.v2.processes.connectors.fsspec.utils import sterilize_dict

CONNECTOR_TYPE = "dropbox"


@dataclass
class DropboxIndexerConfig(FsspecIndexerConfig):
    pass


@dataclass
class DropboxAccessConfig(FsspecAccessConfig):
    token: Optional[str] = None


@dataclass
class DropboxConnectionConfig(FsspecConnectionConfig):
    supported_protocols: list[str] = field(default_factory=lambda: ["dropbox"])
    access_config: DropboxAccessConfig = enhanced_field(
        sensitive=True, default_factory=lambda: DropboxAccessConfig()
    )
    connector_type: str = CONNECTOR_TYPE


@dataclass
class DropboxIndexer(FsspecIndexer):
    connection_config: DropboxConnectionConfig
    index_config: DropboxIndexerConfig
    connector_type: str = CONNECTOR_TYPE

    @requires_dependencies(["dropboxdrivefs", "fsspec"], extras="dropbox")
    def __post_init__(self):
        # dropbox expects the path to start with a /
        if not self.index_config.path_without_protocol.startswith("/"):
            self.index_config.path_without_protocol = "/" + self.index_config.path_without_protocol

    @requires_dependencies(["dropboxdrivefs", "fsspec"], extras="dropbox")
    def run(self, **kwargs: Any) -> Generator[FileData, None, None]:
        return super().run(**kwargs)

    def sterilize_info(self, path) -> dict:
        # the fs.info method defined in the dropboxdrivefs library expects a "url"
        # kwarg rather than "path"; though both refer to the same thing
        info = self.fs.info(url=path)
        return sterilize_dict(data=info)


@dataclass
class DropboxDownloaderConfig(FsspecDownloaderConfig):
    pass


@dataclass
class DropboxDownloader(FsspecDownloader):
    protocol: str = "dropbox"
    connection_config: DropboxConnectionConfig
    connector_type: str = CONNECTOR_TYPE
    download_config: Optional[DropboxDownloaderConfig] = field(
        default_factory=DropboxDownloaderConfig
    )

    @requires_dependencies(["dropboxdrivefs", "fsspec"], extras="dropbox")
    def run(self, file_data: FileData, **kwargs: Any) -> DownloadResponse:
        return super().run(file_data=file_data, **kwargs)

    @requires_dependencies(["dropboxdrivefs", "fsspec"], extras="dropbox")
    async def run_async(self, file_data: FileData, **kwargs: Any) -> DownloadResponse:
        return await super().run_async(file_data=file_data, **kwargs)


@dataclass
class DropboxUploaderConfig(FsspecUploaderConfig):
    pass


@dataclass
class DropboxUploader(FsspecUploader):
    connector_type: str = CONNECTOR_TYPE
    connection_config: DropboxConnectionConfig
    upload_config: DropboxUploaderConfig = field(default=None)

    @requires_dependencies(["dropboxdrivefs", "fsspec"], extras="dropbox")
    def __post_init__(self):
        super().__post_init__()

    @requires_dependencies(["dropboxdrivefs", "fsspec"], extras="dropbox")
    def run(self, contents: list[UploadContent], **kwargs: Any) -> None:
        return super().run(contents=contents, **kwargs)

    @requires_dependencies(["dropboxdrivefs", "fsspec"], extras="dropbox")
    async def run_async(self, path: Path, file_data: FileData, **kwargs: Any) -> None:
        return await super().run_async(path=path, file_data=file_data, **kwargs)


dropbox_source_entry = SourceRegistryEntry(
    indexer=DropboxIndexer,
    indexer_config=DropboxIndexerConfig,
    downloader=DropboxDownloader,
    downloader_config=DropboxDownloaderConfig,
    connection_config=DropboxConnectionConfig,
)

dropbox_destination_entry = DestinationRegistryEntry(
    uploader=DropboxUploader,
    uploader_config=DropboxUploaderConfig,
    connection_config=DropboxConnectionConfig,
)



import logging
from dataclasses import dataclass, field
from typing import Optional, Type

import click

from unstructured_ingest.v2.cli.base.cmd import BaseCmd
from unstructured_ingest.v2.cli.configs import (
    ChunkerCliConfig,
    EmbedderCliConfig,
    PartitionerCliConfig,
    ProcessorCliConfig,
)
from unstructured_ingest.v2.cli.interfaces import CliConfig
from unstructured_ingest.v2.cli.utils import Group, conform_click_options
from unstructured_ingest.v2.logger import logger


@dataclass
class SrcCmd(BaseCmd):
    indexer_config: Optional[Type[CliConfig]] = None
    downloader_config: Optional[Type[CliConfig]] = None
    connection_config: Optional[Type[CliConfig]] = None
    default_configs: list[CliConfig] = field(
        default_factory=lambda: [
            ProcessorCliConfig,
            PartitionerCliConfig,
            EmbedderCliConfig,
            ChunkerCliConfig,
        ]
    )

    def cmd(self, ctx: click.Context, **options) -> None:
        if ctx.invoked_subcommand:
            return

        conform_click_options(options)
        logger.setLevel(logging.DEBUG if options.get("verbose", False) else logging.INFO)
        try:
            pipeline = self.get_pipline(src=self.cmd_name, source_options=options)
            pipeline.run()
        except Exception as e:
            logger.error(f"failed to run source command {self.cmd_name}: {e}", exc_info=True)
            raise click.ClickException(str(e)) from e

    def get_cmd(self) -> click.Group:
        # Dynamically create the command without the use of click decorators
        fn = self.cmd
        fn = click.pass_context(fn)
        cmd = click.group(fn, cls=Group)
        if not isinstance(cmd, click.core.Group):
            raise ValueError(f"generated src command was not of expected type Group: {type(cmd)}")
        cmd.name = self.cli_cmd_name
        cmd.short_help = "v2"
        cmd.invoke_without_command = True
        extras = [
            x for x in [self.indexer_config, self.downloader_config, self.connection_config] if x
        ]
        self.add_options(cmd, extras=extras)

        # TODO remove after v1 no longer supported
        cmd.params.append(
            click.Option(
                ["--output-dir"],
                required=False,
                type=str,
                help="Local path to write partitioned output to",
            )
        )
        return cmd



import enum
import logging
from dataclasses import dataclass, field
from typing import List, Optional, Tuple
from urllib.parse import urlparse
from uuid import UUID

from htmlBuilder.attributes import Style, Type
from htmlBuilder.tags import (
    Body,
    Div,
    Head,
    Html,
    HtmlTag,
    Ol,
    Table,
    Td,
    Th,
    Title,
    Tr,
    Ul,
)
from notion_client.errors import APIResponseError

import unstructured_ingest.connector.notion.types.blocks as notion_blocks
from unstructured_ingest.connector.notion.client import Client
from unstructured_ingest.connector.notion.interfaces import BlockBase
from unstructured_ingest.connector.notion.types.block import Block
from unstructured_ingest.connector.notion.types.database import Database


@dataclass
class TextExtractionResponse:
    text: Optional[str] = None
    child_pages: List[str] = field(default_factory=list)
    child_databases: List[str] = field(default_factory=list)


@dataclass
class HtmlExtractionResponse:
    html: Optional[HtmlTag] = None
    child_pages: List[str] = field(default_factory=list)
    child_databases: List[str] = field(default_factory=list)


def extract_page_html(
    client: Client,
    page_id: str,
    logger: logging.Logger,
) -> HtmlExtractionResponse:
    page_id_uuid = UUID(page_id)
    html_elements: List[Tuple[BlockBase, HtmlTag]] = []
    parent_block: Block = client.blocks.retrieve(block_id=page_id)  # type: ignore
    head = None
    if isinstance(parent_block.block, notion_blocks.ChildPage):
        head = Head([], Title([], parent_block.block.title))
    child_pages: List[str] = []
    child_databases: List[str] = []
    parents: List[Tuple[int, Block]] = [(0, parent_block)]
    processed_block_ids = []
    while len(parents) > 0:
        level, parent = parents.pop(0)
        parent_html = parent.get_html()
        if parent_html:
            html_elements.append((parent.block, parent_html))
        logger.debug(f"processing block: {parent}")
        if isinstance(parent.block, notion_blocks.ChildPage) and parent.id != str(page_id_uuid):
            child_pages.append(parent.id)
            continue
        if isinstance(parent.block, notion_blocks.ChildDatabase):
            child_databases.append(parent.id)
            continue
        if isinstance(parent.block, notion_blocks.Table):
            table_response = build_table(client=client, table=parent)
            html_elements.append((parent.block, table_response.table_html))
            child_pages.extend(table_response.child_pages)
            child_databases.extend(table_response.child_databases)
            continue
        if isinstance(parent.block, notion_blocks.ColumnList):
            column_html = build_columned_list(client=client, column_parent=parent)
            html_elements.append((parent.block, column_html))
            continue
        if isinstance(parent.block, notion_blocks.BulletedListItem):
            bullet_list_resp = build_bulleted_list_children(
                client=client,
                bulleted_list_item_parent=parent,
            )
            if bullet_list_children := bullet_list_resp.child_list:
                html_elements.append((parent.block, bullet_list_children))
            continue
        if isinstance(parent.block, notion_blocks.NumberedListItem):
            numbered_list_resp = build_numbered_list_children(
                client=client,
                numbered_list_item_parent=parent,
            )
            if numbered_list_children := numbered_list_resp.child_list:
                html_elements.append((parent.block, numbered_list_children))
            continue
        if parent.block.can_have_children() and parent.has_children:
            children = []
            for children_block in client.blocks.children.iterate_list(  # type: ignore
                block_id=parent.id,
            ):
                children.extend(children_block)
            if children:
                logger.debug(f"Adding {len(children)} children from parent: {parent}")
                for child in children:
                    if child.id not in processed_block_ids:
                        parents.append((level + 1, child))
        processed_block_ids.append(parent)

    # Join list items
    joined_html_elements = []
    numbered_list_items = []
    bullet_list_items = []
    for block, html in html_elements:
        if isinstance(block, notion_blocks.BulletedListItem):
            bullet_list_items.append(html)
            continue
        if isinstance(block, notion_blocks.NumberedListItem):
            numbered_list_items.append(html)
            continue
        if len(numbered_list_items) > 0:
            joined_html_elements.append(Ol([], numbered_list_items))
            numbered_list_items = []
        if len(bullet_list_items) > 0:
            joined_html_elements.append(Ul([], bullet_list_items))
            bullet_list_items = []
        joined_html_elements.append(html)

    body = Body([], joined_html_elements)
    all_elements = [body]
    if head:
        all_elements = [head] + all_elements
    full_html = Html([], all_elements)
    return HtmlExtractionResponse(
        full_html,
        child_pages=child_pages,
        child_databases=child_databases,
    )


def extract_database_html(
    client: Client,
    database_id: str,
    logger: logging.Logger,
) -> HtmlExtractionResponse:
    logger.debug(f"processing database id: {database_id}")
    database: Database = client.databases.retrieve(database_id=database_id)  # type: ignore
    property_keys = list(database.properties.keys())
    property_keys = sorted(property_keys)
    table_html_rows = []
    child_pages: List[str] = []
    child_databases: List[str] = []
    # Create header row
    table_html_rows.append(Tr([], [Th([], k) for k in property_keys]))

    all_pages = []
    for page_chunk in client.databases.iterate_query(database_id=database_id):  # type: ignore
        all_pages.extend(page_chunk)

    logger.debug(f"Creating {len(all_pages)} rows")
    for page in all_pages:
        if is_database_url(client=client, url=page.url):
            child_databases.append(page.id)
        if is_page_url(client=client, url=page.url):
            child_pages.append(page.id)
        properties = page.properties
        inner_html = [properties.get(k).get_html() for k in property_keys]  # type: ignore
        table_html_rows.append(
            Tr(
                [],
                [Td([], cell) for cell in [html if html else Div([], []) for html in inner_html]],
            ),
        )

    table_html = Table([], table_html_rows)

    return HtmlExtractionResponse(
        html=table_html,
        child_pages=child_pages,
        child_databases=child_databases,
    )


@dataclass
class ChildExtractionResponse:
    child_pages: List[str] = field(default_factory=list)
    child_databases: List[str] = field(default_factory=list)


class QueueEntryType(enum.Enum):
    DATABASE = "database"
    PAGE = "page"


@dataclass
class QueueEntry:
    type: QueueEntryType
    id: UUID


def get_recursive_content_from_page(
    client: Client,
    page_id: str,
    logger: logging.Logger,
) -> ChildExtractionResponse:
    return get_recursive_content(
        client=client,
        init_entry=QueueEntry(type=QueueEntryType.PAGE, id=UUID(page_id)),
        logger=logger,
    )


def get_recursive_content_from_database(
    client: Client,
    database_id: str,
    logger: logging.Logger,
) -> ChildExtractionResponse:
    return get_recursive_content(
        client=client,
        init_entry=QueueEntry(type=QueueEntryType.DATABASE, id=UUID(database_id)),
        logger=logger,
    )


def get_recursive_content(
    client: Client,
    init_entry: QueueEntry,
    logger: logging.Logger,
) -> ChildExtractionResponse:
    parents: List[QueueEntry] = [init_entry]
    child_pages: List[str] = []
    child_dbs: List[str] = []
    processed: List[str] = []
    while len(parents) > 0:
        parent: QueueEntry = parents.pop()
        processed.append(str(parent.id))
        if parent.type == QueueEntryType.PAGE:
            logger.debug(f"Getting child data from page: {parent.id}")
            page_children = []
            try:
                for children_block in client.blocks.children.iterate_list(  # type: ignore
                    block_id=str(parent.id),
                ):
                    page_children.extend(children_block)
            except APIResponseError as api_error:
                logger.error(f"failed to get page with id {parent.id}: {api_error}")
                if str(parent.id) in child_pages:
                    child_pages.remove(str(parent.id))
                continue
            if not page_children:
                continue

            # Extract child pages
            child_pages_from_page = [
                c for c in page_children if isinstance(c.block, notion_blocks.ChildPage)
            ]
            if child_pages_from_page:
                child_page_blocks: List[notion_blocks.ChildPage] = [
                    p.block
                    for p in child_pages_from_page
                    if isinstance(p.block, notion_blocks.ChildPage)
                ]
                logger.debug(
                    "found child pages from parent page {}: {}".format(
                        parent.id,
                        ", ".join([block.title for block in child_page_blocks]),
                    ),
                )
            new_pages = [p.id for p in child_pages_from_page if p.id not in processed]
            new_pages = list(set(new_pages))
            child_pages.extend(new_pages)
            parents.extend(
                [QueueEntry(type=QueueEntryType.PAGE, id=UUID(i)) for i in new_pages],
            )

            # Extract child databases
            child_dbs_from_page = [
                c for c in page_children if isinstance(c.block, notion_blocks.ChildDatabase)
            ]
            if child_dbs_from_page:
                child_db_blocks: List[notion_blocks.ChildDatabase] = [
                    c.block
                    for c in page_children
                    if isinstance(c.block, notion_blocks.ChildDatabase)
                ]
                logger.debug(
                    "found child database from parent page {}: {}".format(
                        parent.id,
                        ", ".join([block.title for block in child_db_blocks]),
                    ),
                )
            new_dbs = [db.id for db in child_dbs_from_page if db.id not in processed]
            new_dbs = list(set(new_dbs))
            child_dbs.extend(new_dbs)
            parents.extend(
                [QueueEntry(type=QueueEntryType.DATABASE, id=UUID(i)) for i in new_dbs],
            )

            linked_to_others: List[notion_blocks.LinkToPage] = [
                c.block for c in page_children if isinstance(c.block, notion_blocks.LinkToPage)
            ]
            for link in linked_to_others:
                if (page_id := link.page_id) and (
                    page_id not in processed and page_id not in child_pages
                ):
                    child_pages.append(page_id)
                    parents.append(QueueEntry(type=QueueEntryType.PAGE, id=UUID(page_id)))
                if (database_id := link.database_id) and (
                    database_id not in processed and database_id not in child_dbs
                ):
                    child_dbs.append(database_id)
                    parents.append(
                        QueueEntry(type=QueueEntryType.DATABASE, id=UUID(database_id)),
                    )

        elif parent.type == QueueEntryType.DATABASE:
            logger.debug(f"Getting child data from database: {parent.id}")
            database_pages = []
            try:
                for page_entries in client.databases.iterate_query(  # type: ignore
                    database_id=str(parent.id),
                ):
                    database_pages.extend(page_entries)
            except APIResponseError as api_error:
                logger.error(f"failed to get database with id {parent.id}: {api_error}")
                if str(parent.id) in child_dbs:
                    child_dbs.remove(str(parent.id))
                continue
            if not database_pages:
                continue

            child_pages_from_db = [
                p for p in database_pages if is_page_url(client=client, url=p.url)
            ]
            if child_pages_from_db:
                logger.debug(
                    "found child pages from parent database {}: {}".format(
                        parent.id,
                        ", ".join([p.url for p in child_pages_from_db]),
                    ),
                )
            new_pages = [p.id for p in child_pages_from_db if p.id not in processed]
            child_pages.extend(new_pages)
            parents.extend(
                [QueueEntry(type=QueueEntryType.PAGE, id=UUID(i)) for i in new_pages],
            )

            child_dbs_from_db = [
                p for p in database_pages if is_database_url(client=client, url=p.url)
            ]
            if child_dbs_from_db:
                logger.debug(
                    "found child database from parent database {}: {}".format(
                        parent.id,
                        ", ".join([db.url for db in child_dbs_from_db]),
                    ),
                )
            new_dbs = [db.id for db in child_dbs_from_db if db.id not in processed]
            child_dbs.extend(new_dbs)
            parents.extend(
                [QueueEntry(type=QueueEntryType.DATABASE, id=UUID(i)) for i in new_dbs],
            )

    return ChildExtractionResponse(
        child_pages=child_pages,
        child_databases=child_dbs,
    )


def is_valid_uuid(uuid_str: str) -> bool:
    try:
        UUID(uuid_str)
        return True
    except Exception:
        return False


def get_uuid_from_url(path: str) -> Optional[str]:
    strings = path.split("-")
    if len(strings) > 0 and is_valid_uuid(strings[-1]):
        return strings[-1]
    return None


def is_page_url(client: Client, url: str):
    parsed_url = urlparse(url)
    path = parsed_url.path.split("/")[-1]
    if parsed_url.netloc != "www.notion.so":
        return False
    page_uuid = get_uuid_from_url(path=path)
    if not page_uuid:
        return False
    check_resp = client.pages.retrieve_status(page_id=page_uuid)
    return check_resp == 200


def is_database_url(client: Client, url: str):
    parsed_url = urlparse(url)
    path = parsed_url.path.split("/")[-1]
    if parsed_url.netloc != "www.notion.so":
        return False
    database_uuid = get_uuid_from_url(path=path)
    if not database_uuid:
        return False
    check_resp = client.databases.retrieve_status(database_id=database_uuid)
    return check_resp == 200


@dataclass
class BuildTableResponse:
    table_html: HtmlTag
    child_pages: List[str] = field(default_factory=list)
    child_databases: List[str] = field(default_factory=list)


def build_table(client: Client, table: Block) -> BuildTableResponse:
    if not isinstance(table.block, notion_blocks.Table):
        raise ValueError(f"block type not table: {type(table.block)}")
    rows: List[notion_blocks.TableRow] = []
    child_pages: List[str] = []
    child_databases: List[str] = []
    for row_chunk in client.blocks.children.iterate_list(  # type: ignore
        block_id=table.id,
    ):
        rows.extend(
            [row.block for row in row_chunk if isinstance(row.block, notion_blocks.TableRow)],
        )

    # Extract child databases and pages
    for row in rows:
        for c in row.cells:
            for rt in c.rich_texts:
                if mention := rt.mention:
                    if mention.type == "page" and (page := mention.page):
                        child_pages.append(page.id)
                    if mention.type == "database" and (database := mention.database):
                        child_databases.append(database.id)

    header: Optional[notion_blocks.TableRow] = None
    if table.block.has_column_header:
        header = rows.pop(0)
    table_html_rows = []
    if header:
        header.is_header = True
        table_html_rows.append(header.get_html())
    table_html_rows.extend([row.get_html() for row in rows])
    html_table = Table([], table_html_rows)

    return BuildTableResponse(
        table_html=html_table,
        child_pages=child_pages,
        child_databases=child_databases,
    )


def build_columned_list(client: Client, column_parent: Block) -> HtmlTag:
    if not isinstance(column_parent.block, notion_blocks.ColumnList):
        raise ValueError(f"block type not column list: {type(column_parent.block)}")
    columns: List[Block] = []
    for column_chunk in client.blocks.children.iterate_list(  # type: ignore
        block_id=column_parent.id,
    ):
        columns.extend(column_chunk)
    num_columns = len(columns)
    columns_content = []
    for column in columns:
        for column_content_chunk in client.blocks.children.iterate_list(  # type: ignore
            block_id=column.id,
        ):
            columns_content.append(
                Div(
                    [Style(f"width:{100/num_columns}%; float: left")],
                    [content.block.get_html() for content in column_content_chunk],
                ),
            )

    return Div([], columns_content)


@dataclass
class BulletedListResponse:
    html: HtmlTag
    child_list: Optional[HtmlTag] = None


bulleted_list_styles = ["circle", "square", "disc"]


def build_bulleted_list_children(
    client: Client,
    bulleted_list_item_parent: Block,
    list_style_ind: int = 0,
) -> BulletedListResponse:
    if not isinstance(bulleted_list_item_parent.block, notion_blocks.BulletedListItem):
        raise ValueError(
            f"block type not bulleted list item: {type(bulleted_list_item_parent.block)}",
        )
    html = bulleted_list_item_parent.get_html()
    if html:
        html.attributes = [Style("margin-left: 10px")]
    if not bulleted_list_item_parent.has_children:
        return BulletedListResponse(
            html=html,
        )
    children = []
    for child_block in client.blocks.children.iterate_list(  # type: ignore
        block_id=bulleted_list_item_parent.id,
    ):
        children.extend(child_block)
    if not children:
        return BulletedListResponse(
            html=bulleted_list_item_parent.get_html(),
        )
    child_html = []
    for child in children:
        child_resp = build_bulleted_list_children(
            client=client,
            bulleted_list_item_parent=child,
            list_style_ind=(list_style_ind + 1) % len(bulleted_list_styles),
        )
        child_html.append(child_resp.html)
        if child_children := child_resp.child_list:
            child_html.append(child_children)

    return BulletedListResponse(
        html=html,
        child_list=Ul(
            [Style(f"list-style-type: {bulleted_list_styles[list_style_ind]}")],
            child_html,
        ),
    )


@dataclass
class NumberedListResponse:
    html: HtmlTag
    child_list: Optional[HtmlTag] = None


numbered_list_types = ["a", "i", "1"]


def build_numbered_list_children(
    client: Client,
    numbered_list_item_parent: Block,
    type_attr_ind=0,
) -> NumberedListResponse:
    if not isinstance(numbered_list_item_parent.block, notion_blocks.NumberedListItem):
        raise ValueError(
            f"block type not numbered list item: {type(numbered_list_item_parent.block)}",
        )
    html = numbered_list_item_parent.get_html()
    if html:
        html.attributes = [Style("margin-left: 10px")]
    if not numbered_list_item_parent.has_children:
        return NumberedListResponse(
            html=html,
        )
    children = []
    for child_block in client.blocks.children.iterate_list(  # type: ignore
        block_id=numbered_list_item_parent.id,
    ):
        children.extend(child_block)
    if not children:
        return NumberedListResponse(
            html=numbered_list_item_parent.get_html(),
        )
    child_html = []
    for child in children:
        child_resp = build_numbered_list_children(
            client=client,
            numbered_list_item_parent=child,
            type_attr_ind=(type_attr_ind + 1) % len(numbered_list_types),
        )
        child_html.append(child_resp.html)
        if child_children := child_resp.child_list:
            child_html.append(child_children)

    return NumberedListResponse(
        html=html,
        child_list=Ol([Type(numbered_list_types[type_attr_ind])], child_html),
    )
